# coding: utf-8
"""
Training module
"""
import argparse
import heapq
import logging
import math
import shutil
import sys
import time
from collections import OrderedDict
from pathlib import Path
from typing import List, Tuple

import numpy as np
import torch
from torch import Tensor
from torch.utils.data import Dataset
from torch.utils.tensorboard import SummaryWriter

from joeynmt.batch import Batch
from joeynmt.builders import build_gradient_clipper, build_optimizer, build_scheduler
from joeynmt.data import load_data, make_data_iter
from joeynmt.helpers import (
    delete_ckpt,
    load_checkpoint,
    load_config,
    log_cfg,
    make_logger,
    make_model_dir,
    parse_train_args,
    set_seed,
    store_attention_plots,
    symlink_update,
    write_list_to_file,
)
from joeynmt.model import Model, _DataParallel, build_model
from joeynmt.prediction import predict, test

# for fp16 training
try:
    from apex import amp

    amp.register_half_function(torch, "einsum")
except ImportError as no_apex:  # noqa: F841
    # error handling in TrainManager object construction
    pass

logger = logging.getLogger(__name__)


class TrainManager:
    """
    Manages training loop, validations, learning rate scheduling
    and early stopping.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, model: Model, cfg: dict) -> None:
        """
        Creates a new TrainManager for a model, specified as in configuration.
        Note: no need to pass batch_class here. see make_data_iter()

        :param model: torch module defining the model
        :param cfg: dictionary containing the training configurations
        """
        (  # pylint: disable=unbalanced-tuple-unpacking
            model_dir,
            load_model,
            load_encoder,
            load_decoder,
            loss_type,
            label_smoothing,
            normalization,
            learning_rate_min,
            keep_best_ckpts,
            logging_freq,
            validation_freq,
            log_valid_sents,
            early_stopping_metric,
            seed,
            shuffle,
            epochs,
            max_updates,
            batch_size,
            batch_type,
            batch_multiplier,
            device,
            n_gpu,
            num_workers,
            fp16,
            reset_best_ckpt,
            reset_scheduler,
            reset_optimizer,
            reset_iter_state,
        ) = parse_train_args(cfg["training"])

        # logging and storing
        self.model_dir = model_dir
        self.tb_writer = SummaryWriter(log_dir=(model_dir / "tensorboard").as_posix())
        self.logging_freq = logging_freq
        self.validation_freq = validation_freq
        self.log_valid_sents = log_valid_sents

        # model
        self.model = model
        self.model.log_parameters_list()
        self.model.loss_function = (loss_type, label_smoothing)
        logger.info(self.model)

        # CPU / GPU
        self.device = device
        self.n_gpu = n_gpu
        self.num_workers = num_workers
        if self.device.type == "cuda":
            self.model.to(self.device)

        # optimization
        self.clip_grad_fun = build_gradient_clipper(config=cfg["training"])
        self.optimizer = build_optimizer(config=cfg["training"],
                                         parameters=self.model.parameters())

        # save/delete checkpoints
        self.num_ckpts = keep_best_ckpts
        self.ckpt_queue: List[Tuple[float, Path]] = []  # heap queue

        # early_stopping
        self.early_stopping_metric = early_stopping_metric
        # early_stopping_metric decides on how to find the early stopping point: ckpts
        # are written when there's a new high/low score for this metric. If we schedule
        # after loss/ppl, we want to minimize the score, else we want to maximize it.
        if self.early_stopping_metric in ["ppl", "loss"]:  # lower is better
            self.minimize_metric = True
        elif self.early_stopping_metric in ["acc", "bleu", "chrf"]:  # higher is better
            self.minimize_metric = False

        # learning rate scheduling
        self.scheduler, self.scheduler_step_at = build_scheduler(
            config=cfg["training"],
            scheduler_mode="min" if self.minimize_metric else "max",
            optimizer=self.optimizer,
            hidden_size=cfg["model"]["encoder"]["hidden_size"],
        )

        # data & batch handling
        self.seed = seed
        self.shuffle = shuffle
        self.epochs = epochs
        self.max_updates = max_updates
        self.max_updates = max_updates
        self.batch_size = batch_size
        self.batch_type = batch_type
        self.learning_rate_min = learning_rate_min
        self.batch_multiplier = batch_multiplier
        self.normalization = normalization

        # Placeholder so that we can use the train_iter in other functions.
        self.train_iter, self.train_iter_state = None, None

        # initialize training statistics
        self.stats = self.TrainStatistics(
            steps=0,
            is_min_lr=False,
            is_max_update=False,
            total_tokens=0,
            best_ckpt_iter=0,
            best_ckpt_score=np.inf if self.minimize_metric else -np.inf,
            minimize_metric=self.minimize_metric,
        )

        # fp16
        self.fp16 = fp16
        if self.fp16:
            if "apex" not in sys.modules:
                # pylint: disable=used-before-assignment
                raise ImportError(
                    "Please install apex from https://www.github.com/nvidia/apex "
                    "to use fp16 training.") from no_apex  # noqa: F821

            self.model, self.optimizer = amp.initialize(self.model,
                                                        self.optimizer,
                                                        opt_level="O1")
            # opt level: one of {"O0", "O1", "O2", "O3"}
            # see https://nvidia.github.io/apex/amp.html#opt-levels

        # load model parameters
        if load_model is not None:
            self.init_from_checkpoint(
                load_model,
                reset_best_ckpt=reset_best_ckpt,
                reset_scheduler=reset_scheduler,
                reset_optimizer=reset_optimizer,
                reset_iter_state=reset_iter_state,
            )
        for layer_name, load_path in [
            ("encoder", load_encoder),
            ("decoder", load_decoder),
        ]:
            if load_path is not None:
                self.init_layers(path=load_path, layer=layer_name)

        # gpu training (should be after apex fp16 initialization)
        if self.n_gpu > 1:
            self.model = _DataParallel(self.model)

        # config for generation
        self.valid_cfg = cfg["testing"].copy()
        self.valid_cfg["beam_size"] = 1  # greedy decoding during train loop
        # in greedy decoding, we use the same batch_size as the one in training
        self.valid_cfg["batch_size"] = self.batch_size
        self.valid_cfg["batch_type"] = self.batch_type
        # no further exploration during training
        self.valid_cfg["n_best"] = 1
        # self.valid_cfg["return_attention"] = False  # don't override this param
        self.valid_cfg["return_prob"] = "none"
        self.valid_cfg["generate_unk"] = True
        self.valid_cfg["repetition_penalty"] = -1  # turn off
        self.valid_cfg["no_repeat_ngram_size"] = -1  # turn off

    def _save_checkpoint(self, new_best: bool, score: float) -> None:
        """
        Save the model's current parameters and the training state to a checkpoint.

        The training state contains the total number of training steps, the total number
        of training tokens, the best checkpoint score and iteration so far, and
        optimizer and scheduler states.

        :param new_best: This boolean signals which symlink we will use for the new
            checkpoint. If it is true, we update best.ckpt.
        :param score: Validation score which is used as key of heap queue. if score is
            float('nan'), the queue won't be updated.
        """
        model_path = Path(self.model_dir) / f"{self.stats.steps}.ckpt"
        model_state_dict = (self.model.module.state_dict() if isinstance(
            self.model, torch.nn.DataParallel) else self.model.state_dict())
        state = {
            "steps":
            self.stats.steps,
            "total_tokens":
            self.stats.total_tokens,
            "best_ckpt_score":
            self.stats.best_ckpt_score,
            "best_ckpt_iteration":
            self.stats.best_ckpt_iter,
            "model_state":
            model_state_dict,
            "optimizer_state":
            self.optimizer.state_dict(),
            "scheduler_state":
            (self.scheduler.state_dict() if self.scheduler is not None else None),
            "amp_state":
            amp.state_dict() if self.fp16 else None,
            "train_iter_state":
            (self.train_iter.batch_sampler.sampler.generator.get_state()),
        }
        torch.save(state, model_path.as_posix())

        # update symlink
        symlink_target = Path(f"{self.stats.steps}.ckpt")
        # last symlink
        last_path = Path(self.model_dir) / "latest.ckpt"
        prev_path = symlink_update(symlink_target, last_path)  # update always
        # best symlink
        best_path = Path(self.model_dir) / "best.ckpt"
        if new_best:
            prev_path = symlink_update(symlink_target, best_path)
            assert best_path.resolve().stem == str(self.stats.best_ckpt_iter)

        # push to and pop from the heap queue
        to_delete = None
        if not math.isnan(score) and self.num_ckpts > 0:
            if len(self.ckpt_queue) < self.num_ckpts:  # no pop, push only
                heapq.heappush(self.ckpt_queue, (score, model_path))
            else:  # push + pop the worst one in the queue
                if self.minimize_metric:
                    # pylint: disable=protected-access
                    heapq._heapify_max(self.ckpt_queue)
                    to_delete = heapq._heappop_max(self.ckpt_queue)
                    heapq.heappush(self.ckpt_queue, (score, model_path))
                    # pylint: enable=protected-access
                else:
                    to_delete = heapq.heappushpop(self.ckpt_queue, (score, model_path))

            if to_delete is not None:
                assert to_delete[1] != model_path  # don't delete the last ckpt
                if to_delete[1].stem != best_path.resolve().stem:
                    delete_ckpt(to_delete[1])  # don't delete the best ckpt

            assert len(self.ckpt_queue) <= self.num_ckpts

            # remove old symlink target if not in queue after push/pop
            if prev_path is not None and prev_path.stem not in [
                    c[1].stem for c in self.ckpt_queue
            ]:
                delete_ckpt(prev_path)

    def init_from_checkpoint(
        self,
        path: Path,
        reset_best_ckpt: bool = False,
        reset_scheduler: bool = False,
        reset_optimizer: bool = False,
        reset_iter_state: bool = False,
    ) -> None:
        """
        Initialize the trainer from a given checkpoint file.

        This checkpoint file contains not only model parameters, but also
        scheduler and optimizer states, see `self._save_checkpoint`.

        :param path: path to checkpoint
        :param reset_best_ckpt: reset tracking of the best checkpoint,
                                use for domain adaptation with a new dev
                                set or when using a new metric for fine-tuning.
        :param reset_scheduler: reset the learning rate scheduler, and do not
                                use the one stored in the checkpoint.
        :param reset_optimizer: reset the optimizer, and do not use the one
                                stored in the checkpoint.
        :param reset_iter_state: reset the sampler's internal state and do not
                                use the one stored in the checkpoint.
        """
        logger.info("Loading model from %s", path)
        model_checkpoint = load_checkpoint(path=path, device=self.device)

        # restore model and optimizer parameters
        self.model.load_state_dict(model_checkpoint["model_state"])

        if not reset_optimizer:
            self.optimizer.load_state_dict(model_checkpoint["optimizer_state"])
        else:
            logger.info("Reset optimizer.")

        if not reset_scheduler:
            if (model_checkpoint["scheduler_state"] is not None
                    and self.scheduler is not None):
                self.scheduler.load_state_dict(model_checkpoint["scheduler_state"])
        else:
            logger.info("Reset scheduler.")

        if not reset_best_ckpt:
            self.stats.best_ckpt_score = model_checkpoint["best_ckpt_score"]
            self.stats.best_ckpt_iter = model_checkpoint["best_ckpt_iteration"]
        else:
            logger.info("Reset tracking of the best checkpoint.")

        if not reset_iter_state:
            # restore counters
            assert "train_iter_state" in model_checkpoint
            self.stats.steps = model_checkpoint["steps"]
            self.stats.total_tokens = model_checkpoint["total_tokens"]
            self.train_iter_state = model_checkpoint["train_iter_state"]
        else:
            # reset counters if explicitly 'train_iter_state: True' in config
            logger.info("Reset data iterator (random seed: {%d}).", self.seed)

        # move to gpu
        if self.device.type == "cuda":
            self.model.to(self.device)

        # fp16
        if self.fp16 and model_checkpoint.get("amp_state", None) is not None:
            amp.load_state_dict(model_checkpoint["amp_state"])

    def init_layers(self, path: Path, layer: str) -> None:
        """
        Initialize encoder decoder layers from a given checkpoint file.

        :param path: path to checkpoint
        :param layer: layer name; 'encoder' or 'decoder' expected
        """
        assert path is not None
        layer_state_dict = OrderedDict()
        logger.info("Loading %s laysers from %s", layer, path)
        ckpt = load_checkpoint(path=path, device=self.device)
        for k, v in ckpt["model_state"].items():
            if k.startswith(layer):
                layer_state_dict[k] = v
        self.model.load_state_dict(layer_state_dict, strict=False)

    def train_and_validate(self, train_data: Dataset, valid_data: Dataset) -> None:
        """
        Train the model and validate it from time to time on the validation set.

        :param train_data: training data
        :param valid_data: validation data
        """
        # pylint: disable=too-many-branches,too-many-statements
        self.train_iter = make_data_iter(
            dataset=train_data,
            batch_size=self.batch_size,
            batch_type=self.batch_type,
            seed=self.seed,
            shuffle=self.shuffle,
            num_workers=self.num_workers,
            device=self.device,
            pad_index=self.model.pad_index,
        )

        if self.train_iter_state is not None:
            self.train_iter.batch_sampler.sampler.generator.set_state(
                self.train_iter_state.cpu())

        #################################################################
        # simplify accumulation logic:
        #################################################################
        # for epoch in range(epochs):
        #     self.model.zero_grad()
        #     epoch_loss = 0.0
        #     batch_loss = 0.0
        #     for i, batch in enumerate(self.train_iter):
        #
        #         # gradient accumulation:
        #         # loss.backward() inside _train_step()
        #         batch_loss += self._train_step(inputs)
        #
        #         if (i + 1) % self.batch_multiplier == 0:
        #             self.optimizer.step()     # update!
        #             self.model.zero_grad()    # reset gradients
        #             self.steps += 1           # increment counter
        #
        #             epoch_loss += batch_loss  # accumulate batch loss
        #             batch_loss = 0            # reset batch loss
        #
        #     # leftovers are just ignored.
        #################################################################

        logger.info(
            "Train stats:\n"
            "\tdevice: %s\n"
            "\tn_gpu: %d\n"
            "\t16-bits training: %r\n"
            "\tgradient accumulation: %d\n"
            "\tbatch size per device: %d\n"
            "\teffective batch size (w. parallel & accumulation): %d",
            self.device.type,  # next(self.model.parameters()).device
            self.n_gpu,
            self.fp16,
            self.batch_multiplier,
            self.batch_size // self.n_gpu if self.n_gpu > 1 else self.batch_size,
            self.batch_size * self.batch_multiplier,
        )

        try:
            for epoch_no in range(self.epochs):
                logger.info("EPOCH %d", epoch_no + 1)

                if self.scheduler_step_at == "epoch":
                    self.scheduler.step(epoch=epoch_no)

                self.model.train()

                # Reset statistics for each epoch.
                start = time.time()
                total_valid_duration = 0
                start_tokens = self.stats.total_tokens
                self.model.zero_grad()
                epoch_loss = 0
                total_batch_loss = 0
                total_n_correct = 0

                # subsample train data each epoch
                if train_data.random_subset > 0:
                    try:
                        train_data.reset_random_subset()
                        train_data.sample_random_subset(seed=epoch_no)
                        logger.info(
                            "Sample random subset from dev set: n=%d, seed=%d",
                            len(train_data),
                            epoch_no,
                        )
                    except AssertionError as e:
                        logger.warning(e)

                batch: Batch  # yield a joeynmt Batch object
                for i, batch in enumerate(self.train_iter):
                    # sort batch now by src length and keep track of order
                    batch.sort_by_src_length()

                    # get batch loss
                    norm_batch_loss, n_correct = self._train_step(batch)
                    total_batch_loss += norm_batch_loss
                    total_n_correct += n_correct

                    # update!
                    if (i + 1) % self.batch_multiplier == 0:
                        # clip gradients (in-place)
                        if self.clip_grad_fun is not None:
                            if self.fp16:
                                self.clip_grad_fun(
                                    parameters=amp.master_params(self.optimizer))
                            else:
                                self.clip_grad_fun(parameters=self.model.parameters())

                        # make gradient step
                        self.optimizer.step()

                        # decay lr
                        if self.scheduler_step_at == "step":
                            self.scheduler.step(self.stats.steps)

                        # reset gradients
                        self.model.zero_grad()

                        # increment step counter
                        self.stats.steps += 1
                        if self.stats.steps >= self.max_updates:
                            self.stats.is_max_update = True

                        # log learning progress
                        if self.stats.steps % self.logging_freq == 0:
                            elapsed = time.time() - start - total_valid_duration
                            elapsed_tok = self.stats.total_tokens - start_tokens
                            token_accuracy = total_n_correct / elapsed_tok
                            self.tb_writer.add_scalar("train/batch_loss",
                                                      total_batch_loss,
                                                      self.stats.steps)
                            self.tb_writer.add_scalar("train/batch_acc", token_accuracy,
                                                      self.stats.steps)
                            logger.info(
                                "Epoch %3d, "
                                "Step: %8d, "
                                "Batch Loss: %12.6f, "
                                "Batch Acc: %.6f, "
                                "Tokens per Sec: %8.0f, "
                                "Lr: %.6f",
                                epoch_no + 1,
                                self.stats.steps,
                                total_batch_loss,
                                token_accuracy,
                                elapsed_tok / elapsed,
                                self.optimizer.param_groups[0]["lr"],
                            )
                            start = time.time()
                            total_valid_duration = 0
                            start_tokens = self.stats.total_tokens

                        # update epoch_loss
                        epoch_loss += total_batch_loss  # accumulate loss
                        total_batch_loss = 0  # reset batch loss
                        total_n_correct = 0  # reset batch accuracy

                        # validate on the entire dev set
                        if self.stats.steps % self.validation_freq == 0:
                            valid_duration = self._validate(valid_data)
                            total_valid_duration += valid_duration

                        # check current_lr
                        current_lr = self.optimizer.param_groups[0]["lr"]
                        if current_lr < self.learning_rate_min:
                            self.stats.is_min_lr = True

                        self.tb_writer.add_scalar("train/learning_rate", current_lr,
                                                  self.stats.steps)

                    if self.stats.is_min_lr or self.stats.is_max_update:
                        break

                if self.stats.is_min_lr or self.stats.is_max_update:
                    log_str = (f"minimum lr {self.learning_rate_min}"
                               if self.stats.is_min_lr else
                               f"maximum num. of updates {self.max_updates}")
                    logger.info("Training ended since %s was reached.", log_str)
                    break

                logger.info(
                    "Epoch %3d: total training loss %.2f",
                    epoch_no + 1,
                    epoch_loss,
                )
            else:
                logger.info("Training ended after %3d epochs.", epoch_no + 1)
            logger.info(
                "Best validation result (greedy) "
                "at step %8d: %6.2f %s.",
                self.stats.best_ckpt_iter,
                self.stats.best_ckpt_score,
                self.early_stopping_metric,
            )
        except KeyboardInterrupt:
            self._save_checkpoint(False, float("nan"))

        self.tb_writer.close()  # close Tensorboard writer

    def _train_step(self, batch: Batch) -> Tensor:
        """
        Train the model on one batch: Compute the loss.

        :param batch: training batch
        :return:
            - losses for batch (sum)
            - number of correct tokens for batch (sum)
        """
        # reactivate training
        self.model.train()

        # get loss (run as during training with teacher forcing)
        batch_loss, _, _, correct_tokens = self.model(return_type="loss", **vars(batch))

        # normalize batch loss
        norm_batch_loss = batch.normalize(
            batch_loss,
            normalization=self.normalization,
            n_gpu=self.n_gpu,
            n_accumulation=self.batch_multiplier,
        )

        # sum over multiple gpus
        n_correct_tokens = batch.normalize(correct_tokens, "sum", self.n_gpu)

        # accumulate gradients
        if self.fp16:
            with amp.scale_loss(norm_batch_loss, self.optimizer) as scaled_loss:
                scaled_loss.backward()
        else:
            norm_batch_loss.backward()

        # increment token counter
        self.stats.total_tokens += batch.ntokens

        return norm_batch_loss.item(), n_correct_tokens.item()

    def _validate(self, valid_data: Dataset):
        if valid_data.random_subset > 0:  # subsample validation set each valid step
            try:
                valid_data.reset_random_subset()
                valid_data.sample_random_subset(seed=self.stats.steps)
                logger.info(
                    "Sample random subset from dev set: n=%d, seed=%d",
                    len(valid_data),
                    self.stats.steps,
                )
            except AssertionError as e:
                logger.warning(e)

        valid_start_time = time.time()
        (
            valid_scores,
            valid_references,
            valid_hypotheses,
            valid_hypotheses_raw,
            valid_sequence_scores,  # pylint: disable=unused-variable
            valid_attention_scores,
        ) = predict(
            model=self.model,
            data=valid_data,
            compute_loss=True,
            device=self.device,
            n_gpu=self.n_gpu,
            normalization=self.normalization,
            cfg=self.valid_cfg,
        )
        valid_duration = time.time() - valid_start_time

        # for eval_metric in ['loss', 'ppl', 'acc'] + self.eval_metrics:
        for eval_metric, score in valid_scores.items():
            if not math.isnan(score):
                self.tb_writer.add_scalar(f"valid/{eval_metric}", score,
                                          self.stats.steps)

        ckpt_score = valid_scores[self.early_stopping_metric]

        if self.scheduler_step_at == "validation":
            self.scheduler.step(metrics=ckpt_score)

        # update new best
        new_best = self.stats.is_best(ckpt_score)
        if new_best:
            self.stats.best_ckpt_score = ckpt_score
            self.stats.best_ckpt_iter = self.stats.steps
            logger.info(
                "Hooray! New best validation result [%s]!",
                self.early_stopping_metric,
            )

        # save checkpoints
        is_better = (self.stats.is_better(ckpt_score, self.ckpt_queue)
                     if len(self.ckpt_queue) > 0 else True)
        if self.num_ckpts < 0 or is_better:
            self._save_checkpoint(new_best, ckpt_score)

        # append to validation report
        self._add_report(valid_scores=valid_scores, new_best=new_best)

        self._log_examples(
            references=valid_references,
            hypotheses=valid_hypotheses,
            hypotheses_raw=valid_hypotheses_raw,
            data=valid_data,
        )

        # store validation set outputs
        write_list_to_file(self.model_dir / f"{self.stats.steps}.hyps",
                           valid_hypotheses)

        # store attention plots for selected valid sentences
        if valid_attention_scores:
            store_attention_plots(
                attentions=valid_attention_scores,
                targets=valid_hypotheses_raw,
                sources=valid_data.get_list(lang=valid_data.src_lang, tokenized=True),
                indices=self.log_valid_sents,
                output_prefix=(self.model_dir / f"att.{self.stats.steps}").as_posix(),
                tb_writer=self.tb_writer,
                steps=self.stats.steps,
            )

        return valid_duration

    def _add_report(self, valid_scores: dict, new_best: bool = False) -> None:
        """
        Append a one-line report to validation logging file.

        :param valid_scores: validation evaluation score [eval_metric]
        :param new_best: whether this is a new best model
        """
        current_lr = self.optimizer.param_groups[0]["lr"]

        valid_file = self.model_dir / "validations.txt"
        with valid_file.open("a", encoding="utf-8") as opened_file:
            score_str = "\t".join([f"Steps: {self.stats.steps}"] + [
                f"{eval_metric}: {score:.5f}"
                for eval_metric, score in valid_scores.items() if not math.isnan(score)
            ] + [f"LR: {current_lr:.8f}", "*" if new_best else ""])
            opened_file.write(f"{score_str}\n")

    def _log_examples(
        self,
        hypotheses: List[str],
        references: List[str],
        hypotheses_raw: List[List[str]],
        data: Dataset,
    ) -> None:
        """
        Log the first `self.log_valid_sents` sentences from given examples.

        :param hypotheses: decoded hypotheses (list of strings)
        :param references: decoded references (list of strings)
        :param hypotheses_raw: raw hypotheses (list of list of tokens)
        :param data: Dataset
        """
        for p in self.log_valid_sents:
            if p >= len(hypotheses):
                continue
            logger.info("Example #%d", p)

            # tokenized text
            tokenized_src = data.get_item(idx=p, lang=data.src_lang)
            tokenized_trg = data.get_item(idx=p, lang=data.trg_lang)
            logger.debug("\tTokenized source:     %s", tokenized_src)
            logger.debug("\tTokenized reference:  %s", tokenized_trg)
            logger.debug("\tTokenized hypothesis: %s", hypotheses_raw[p])

            # detokenized text
            logger.info("\tSource:     %s", data.src[p])
            logger.info("\tReference:  %s", references[p])
            logger.info("\tHypothesis: %s", hypotheses[p])

    class TrainStatistics:

        def __init__(
            self,
            steps: int = 0,
            is_min_lr: bool = False,
            is_max_update: bool = False,
            total_tokens: int = 0,
            best_ckpt_iter: int = 0,
            best_ckpt_score: float = np.inf,
            minimize_metric: bool = True,
        ) -> None:
            self.steps = steps  # global update step counter
            self.is_min_lr = is_min_lr  # stop by reaching learning rate minimum
            self.is_max_update = is_max_update  # stop by reaching max num of updates
            self.total_tokens = total_tokens  # number of total tokens seen so far
            self.best_ckpt_iter = best_ckpt_iter  # store iteration point of best ckpt
            self.best_ckpt_score = best_ckpt_score  # initial values for best scores
            self.minimize_metric = minimize_metric  # minimize or maximize score

        def is_best(self, score):
            if self.minimize_metric:
                is_best = score < self.best_ckpt_score
            else:
                is_best = score > self.best_ckpt_score
            return is_best

        def is_better(self, score: float, heap_queue: list):
            assert len(heap_queue) > 0
            if self.minimize_metric:
                is_better = score < heapq.nlargest(1, heap_queue)[0][0]
            else:
                is_better = score > heapq.nsmallest(1, heap_queue)[0][0]
            return is_better


def train(cfg_file: str, skip_test: bool = False) -> None:
    """
    Main training function. After training, also test on test data if given.

    :param cfg_file: path to configuration yaml file
    :param skip_test: whether a test should be run or not after training
    """
    # read config file
    cfg = load_config(Path(cfg_file))

    # make logger
    model_dir = make_model_dir(
        Path(cfg["training"]["model_dir"]),
        overwrite=cfg["training"].get("overwrite", False),
    )
    joeynmt_version = make_logger(model_dir, mode="train")
    if "joeynmt_version" in cfg:
        assert str(joeynmt_version) == str(cfg["joeynmt_version"]), (
            f"You are using JoeyNMT version {joeynmt_version}, "
            f'but {cfg["joeynmt_version"]} is expected in the given config.')
    # TODO: save version number in model checkpoints

    # write all entries of config to the log
    log_cfg(cfg)

    # store copy of original training config in model dir
    shutil.copy2(cfg_file, (model_dir / "config.yaml").as_posix())

    # set the random seed
    set_seed(seed=cfg["training"].get("random_seed", 42))

    # load the data
    src_vocab, trg_vocab, train_data, dev_data, test_data = load_data(
        data_cfg=cfg["data"])

    # store the vocabs and tokenizers
    src_vocab.to_file(model_dir / "src_vocab.txt")
    if hasattr(train_data.tokenizer[train_data.src_lang], "copy_cfg_file"):
        train_data.tokenizer[train_data.src_lang].copy_cfg_file(model_dir)
    trg_vocab.to_file(model_dir / "trg_vocab.txt")
    if hasattr(train_data.tokenizer[train_data.trg_lang], "copy_cfg_file"):
        train_data.tokenizer[train_data.trg_lang].copy_cfg_file(model_dir)

    # build an encoder-decoder model
    model = build_model(cfg["model"], src_vocab=src_vocab, trg_vocab=trg_vocab)

    # for training management, e.g. early stopping and model selection
    trainer = TrainManager(model=model, cfg=cfg)

    # train the model
    trainer.train_and_validate(train_data=train_data, valid_data=dev_data)

    if not skip_test:
        # predict with the best model on validation and test
        # (if test data is available)

        ckpt = model_dir / f"{trainer.stats.best_ckpt_iter}.ckpt"
        output_path = model_dir / f"{trainer.stats.best_ckpt_iter:08d}.hyps"

        datasets_to_test = {
            "dev": dev_data,
            "test": test_data,
            "src_vocab": src_vocab,
            "trg_vocab": trg_vocab,
        }
        test(
            cfg_file,
            ckpt=ckpt.as_posix(),
            output_path=output_path.as_posix(),
            datasets=datasets_to_test,
        )
    else:
        logger.info("Skipping test after training")


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Joey-NMT")
    parser.add_argument(
        "config",
        default="configs/default.yaml",
        type=str,
        help="Training configuration file (yaml).",
    )
    args = parser.parse_args()
    train(cfg_file=args.config)



# coding: utf-8
"""
Attention modules
"""
from typing import Tuple

import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor, nn


class AttentionMechanism(nn.Module):
    """
    Base attention class
    """

    def forward(self, *inputs):
        raise NotImplementedError("Implement this.")


class BahdanauAttention(AttentionMechanism):
    """
    Implements Bahdanau (MLP) attention

    Section A.1.2 in https://arxiv.org/abs/1409.0473.
    """

    def __init__(self, hidden_size: int = 1, key_size: int = 1, query_size: int = 1):
        """
        Creates attention mechanism.

        :param hidden_size: size of the projection for query and key
        :param key_size: size of the attention input keys
        :param query_size: size of the query
        """

        super().__init__()

        self.key_layer = nn.Linear(key_size, hidden_size, bias=False)
        self.query_layer = nn.Linear(query_size, hidden_size, bias=False)
        self.energy_layer = nn.Linear(hidden_size, 1, bias=False)

        self.proj_keys = None  # to store projected keys
        self.proj_query = None  # projected query

    def forward(self, query: Tensor, mask: Tensor,
                values: Tensor) -> Tuple[Tensor, Tensor]:
        """
        Bahdanau MLP attention forward pass.

        :param query: the item (decoder state) to compare with the keys/memory,
            shape (batch_size, 1, decoder.hidden_size)
        :param mask: mask out keys position (0 in invalid positions, 1 else),
            shape (batch_size, 1, src_length)
        :param values: values (encoder states),
            shape (batch_size, src_length, encoder.hidden_size)
        :return:
            - context vector of shape (batch_size, 1, value_size),
            - attention probabilities of shape (batch_size, 1, src_length)
        """
        # pylint: disable=arguments-differ
        self._check_input_shapes_forward(query=query, mask=mask, values=values)

        assert mask is not None, "mask is required"
        assert self.proj_keys is not None, "projection keys have to get pre-computed"

        # We first project the query (the decoder state).
        # The projected keys (the encoder states) were already pre-computed.
        self.compute_proj_query(query)

        # Calculate scores.
        # proj_keys: batch x src_len x hidden_size
        # proj_query: batch x 1 x hidden_size
        scores = self.energy_layer(torch.tanh(self.proj_query + self.proj_keys))
        # scores: batch x src_len x 1

        scores = scores.squeeze(2).unsqueeze(1)
        # scores: batch x 1 x time

        # mask out invalid positions by filling the masked out parts with -inf
        scores = torch.where(mask > 0, scores, scores.new_full([1], -np.inf))

        # turn scores to probabilities
        alphas = F.softmax(scores, dim=-1)  # batch x 1 x time

        # the context vector is the weighted sum of the values
        context = alphas @ values  # batch x 1 x value_size

        return context, alphas

    def compute_proj_keys(self, keys: Tensor) -> None:
        """
        Compute the projection of the keys.
        Is efficient if pre-computed before receiving individual queries.

        :param keys:
        :return:
        """
        self.proj_keys = self.key_layer(keys)

    def compute_proj_query(self, query: Tensor):
        """
        Compute the projection of the query.

        :param query:
        :return:
        """
        self.proj_query = self.query_layer(query)

    def _check_input_shapes_forward(self, query: Tensor, mask: Tensor,
                                    values: Tensor) -> None:
        """
        Make sure that inputs to `self.forward` are of correct shape.
        Same input semantics as for `self.forward`.

        :param query:
        :param mask:
        :param values:
        :return:
        """
        assert query.shape[0] == values.shape[0] == mask.shape[0]
        assert query.shape[1] == 1 == mask.shape[1]
        assert query.shape[2] == self.query_layer.in_features
        assert values.shape[2] == self.key_layer.in_features
        assert mask.shape[2] == values.shape[1]

    def __repr__(self):
        return "BahdanauAttention"


class LuongAttention(AttentionMechanism):
    """
    Implements Luong (bilinear / multiplicative) attention.

    Eq. 8 ("general") in http://aclweb.org/anthology/D15-1166.
    """

    def __init__(self, hidden_size: int = 1, key_size: int = 1):
        """
        Creates attention mechanism.

        :param hidden_size: size of the key projection layer, has to be equal
            to decoder hidden size
        :param key_size: size of the attention input keys
        """

        super().__init__()
        self.key_layer = nn.Linear(in_features=key_size,
                                   out_features=hidden_size,
                                   bias=False)
        self.proj_keys = None  # projected keys

    def forward(self, query: Tensor, mask: Tensor,
                values: Tensor) -> Tuple[Tensor, Tensor]:
        """
        Luong (multiplicative / bilinear) attention forward pass.
        Computes context vectors and attention scores for a given query and
        all masked values and returns them.

        :param query: the item (decoder state) to compare with the keys/memory,
            shape (batch_size, 1, decoder.hidden_size)
        :param mask: mask out keys position (0 in invalid positions, 1 else),
            shape (batch_size, 1, src_length)
        :param values: values (encoder states),
            shape (batch_size, src_length, encoder.hidden_size)
        :return:
            - context vector of shape (batch_size, 1, value_size),
            - attention probabilities of shape (batch_size, 1, src_length)
        """
        # pylint: disable=arguments-differ
        self._check_input_shapes_forward(query=query, mask=mask, values=values)

        assert self.proj_keys is not None, "projection keys have to get pre-computed"
        assert mask is not None, "mask is required"

        # scores: batch_size x 1 x src_length
        scores = query @ self.proj_keys.transpose(1, 2)

        # mask out invalid positions by filling the masked out parts with -inf
        scores = torch.where(mask > 0, scores, scores.new_full([1], -np.inf))

        # turn scores to probabilities
        alphas = F.softmax(scores, dim=-1)  # batch x 1 x src_len

        # the context vector is the weighted sum of the values
        context = alphas @ values  # batch x 1 x values_size

        return context, alphas

    def compute_proj_keys(self, keys: Tensor) -> None:
        """
        Compute the projection of the keys and assign them to `self.proj_keys`.
        This pre-computation is efficiently done for all keys
        before receiving individual queries.

        :param keys: shape (batch_size, src_length, encoder.hidden_size)
        """
        # proj_keys: batch x src_len x hidden_size
        self.proj_keys = self.key_layer(keys)

    def _check_input_shapes_forward(self, query: Tensor, mask: Tensor,
                                    values: Tensor) -> None:
        """
        Make sure that inputs to `self.forward` are of correct shape.
        Same input semantics as for `self.forward`.

        :param query:
        :param mask:
        :param values:
        :return:
        """
        assert query.shape[0] == values.shape[0] == mask.shape[0]
        assert query.shape[1] == 1 == mask.shape[1]
        assert query.shape[2] == self.key_layer.out_features
        assert values.shape[2] == self.key_layer.in_features
        assert mask.shape[2] == values.shape[1]

    def __repr__(self):
        return "LuongAttention"



# coding: utf-8
"""
Module to represents whole models
"""
import logging
from pathlib import Path
from typing import Tuple

import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor, nn

from joeynmt.decoders import Decoder, RecurrentDecoder, TransformerDecoder
from joeynmt.embeddings import Embeddings
from joeynmt.encoders import Encoder, RecurrentEncoder, TransformerEncoder
from joeynmt.helpers import ConfigurationError
from joeynmt.initialization import initialize_model
from joeynmt.loss import XentLoss
from joeynmt.vocabulary import Vocabulary

logger = logging.getLogger(__name__)


class Model(nn.Module):
    """
    Base Model class
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        encoder: Encoder,
        decoder: Decoder,
        src_embed: Embeddings,
        trg_embed: Embeddings,
        src_vocab: Vocabulary,
        trg_vocab: Vocabulary,
    ) -> None:
        """
        Create a new encoder-decoder model

        :param encoder: encoder
        :param decoder: decoder
        :param src_embed: source embedding
        :param trg_embed: target embedding
        :param src_vocab: source vocabulary
        :param trg_vocab: target vocabulary
        """
        super().__init__()

        self.src_embed = src_embed
        self.trg_embed = trg_embed
        self.encoder = encoder
        self.decoder = decoder
        self.src_vocab = src_vocab
        self.trg_vocab = trg_vocab
        self.pad_index = self.trg_vocab.pad_index
        self.bos_index = self.trg_vocab.bos_index
        self.eos_index = self.trg_vocab.eos_index
        self.unk_index = self.trg_vocab.unk_index
        self._loss_function = None  # set by the TrainManager

    @property
    def loss_function(self):
        return self._loss_function

    @loss_function.setter
    def loss_function(self, cfg: Tuple):
        loss_type, label_smoothing = cfg
        assert loss_type == "crossentropy"
        self._loss_function = XentLoss(pad_index=self.pad_index,
                                       smoothing=label_smoothing)

    def forward(self,
                return_type: str = None,
                **kwargs) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        """Interface for multi-gpu

        For DataParallel, We need to encapsulate all model call: `model.encode()`,
        `model.decode()`, and `model.encode_decode()` by `model.__call__()`.
        `model.__call__()` triggers model.forward() together with pre hooks and post
        hooks, which takes care of multi-gpu distribution.

        :param return_type: one of {"loss", "encode", "decode"}
        """
        if return_type is None:
            raise ValueError("Please specify return_type: "
                             "{`loss`, `encode`, `decode`}.")

        if return_type == "loss":
            assert self.loss_function is not None
            assert "trg" in kwargs and "trg_mask" in kwargs  # need trg to compute loss

            out, _, _, _ = self._encode_decode(**kwargs)

            # compute log probs
            log_probs = F.log_softmax(out, dim=-1)

            # compute batch loss
            # pylint: disable=not-callable
            batch_loss = self.loss_function(log_probs, **kwargs)

            # count correct tokens before decoding (for accuracy)
            trg_mask = kwargs["trg_mask"].squeeze(1)
            assert kwargs["trg"].size() == trg_mask.size()
            n_correct = torch.sum(
                log_probs.argmax(-1).masked_select(trg_mask).eq(
                    kwargs["trg"].masked_select(trg_mask)))

            # return batch loss
            #     = sum over all elements in batch that are not pad
            return_tuple = (batch_loss, log_probs, None, n_correct)

        elif return_type == "encode":
            kwargs["pad"] = True  # TODO: only if multi-gpu
            encoder_output, encoder_hidden = self._encode(**kwargs)

            # return encoder outputs
            return_tuple = (encoder_output, encoder_hidden, None, None)

        elif return_type == "decode":
            outputs, hidden, att_probs, att_vectors = self._decode(**kwargs)

            # return decoder outputs
            return_tuple = (outputs, hidden, att_probs, att_vectors)

        return tuple(return_tuple)

    def _encode_decode(
        self,
        src: Tensor,
        trg_input: Tensor,
        src_mask: Tensor,
        src_length: Tensor,
        trg_mask: Tensor = None,
        **kwargs,
    ) -> Tensor:
        """
        First encodes the source sentence.
        Then produces the target one word at a time.

        :param src: source input
        :param trg_input: target input
        :param src_mask: source mask
        :param src_length: length of source inputs
        :param trg_mask: target mask
        :return: decoder outputs
        """
        encoder_output, encoder_hidden = self._encode(src=src,
                                                      src_length=src_length,
                                                      src_mask=src_mask,
                                                      **kwargs)

        unroll_steps = trg_input.size(1)

        return self._decode(
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            src_mask=src_mask,
            trg_input=trg_input,
            unroll_steps=unroll_steps,
            trg_mask=trg_mask,
            **kwargs,
        )

    def _encode(self, src: Tensor, src_length: Tensor, src_mask: Tensor,
                **_kwargs) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Encodes the source sentence.

        :param src:
        :param src_length:
        :param src_mask:
        :return:
            - encoder_outputs
            - hidden_concat
            - src_mask
        """
        src_input = src if self.src_embed is None else self.src_embed(src)
        return self.encoder(src_input, src_length, src_mask, **_kwargs)

    def _decode(
        self,
        encoder_output: Tensor,
        encoder_hidden: Tensor,
        src_mask: Tensor,
        trg_input: Tensor,
        unroll_steps: int,
        decoder_hidden: Tensor = None,
        att_vector: Tensor = None,
        trg_mask: Tensor = None,
        **_kwargs,
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
        """
        Decode, given an encoded source sentence.

        :param encoder_output: encoder states for attention computation
        :param encoder_hidden: last encoder state for decoder initialization
        :param src_mask: source mask, 1 at valid tokens
        :param trg_input: target inputs
        :param unroll_steps: number of steps to unrol the decoder for
        :param decoder_hidden: decoder hidden state (optional)
        :param att_vector: previous attention vector (optional)
        :param trg_mask: mask for target steps
        :return: decoder outputs
            - decoder_output
            - decoder_hidden
            - att_prob
            - att_vector
        """
        return self.decoder(
            trg_embed=self.trg_embed(trg_input),
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            src_mask=src_mask,
            unroll_steps=unroll_steps,
            hidden=decoder_hidden,
            prev_att_vector=att_vector,
            trg_mask=trg_mask,
            **_kwargs,
        )

    def __repr__(self) -> str:
        """
        String representation: a description of encoder, decoder and embeddings

        :return: string representation
        """
        return (f"{self.__class__.__name__}(\n"
                f"\tencoder={self.encoder},\n"
                f"\tdecoder={self.decoder},\n"
                f"\tsrc_embed={self.src_embed},\n"
                f"\ttrg_embed={self.trg_embed},\n"
                f"\tloss_function={self.loss_function})")

    def log_parameters_list(self) -> None:
        """
        Write all model parameters (name, shape) to the log.
        """
        model_parameters = filter(lambda p: p.requires_grad, self.parameters())
        n_params = sum([np.prod(p.size()) for p in model_parameters])
        logger.info("Total params: %d", n_params)
        trainable_params = [n for (n, p) in self.named_parameters() if p.requires_grad]
        logger.debug("Trainable parameters: %s", sorted(trainable_params))
        assert trainable_params


class _DataParallel(nn.DataParallel):
    """DataParallel wrapper to pass through the model attributes"""

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            return getattr(self.module, name)


def build_model(cfg: dict = None,
                src_vocab: Vocabulary = None,
                trg_vocab: Vocabulary = None) -> Model:
    """
    Build and initialize the model according to the configuration.

    :param cfg: dictionary configuration containing model specifications
    :param src_vocab: source vocabulary
    :param trg_vocab: target vocabulary
    :return: built and initialized model
    """
    logger.info("Building an encoder-decoder model...")
    enc_cfg = cfg["encoder"]
    dec_cfg = cfg["decoder"]

    src_pad_index = src_vocab.pad_index
    trg_pad_index = trg_vocab.pad_index

    src_embed = Embeddings(
        **enc_cfg["embeddings"],
        vocab_size=len(src_vocab),
        padding_idx=src_pad_index,
    )

    # this ties source and target embeddings for softmax layer tying, see further below
    if cfg.get("tied_embeddings", False):
        if src_vocab == trg_vocab:
            trg_embed = src_embed  # share embeddings for src and trg
        else:
            raise ConfigurationError(
                "Embedding cannot be tied since vocabularies differ.")
    else:
        trg_embed = Embeddings(
            **dec_cfg["embeddings"],
            vocab_size=len(trg_vocab),
            padding_idx=trg_pad_index,
        )

    # build encoder
    enc_dropout = enc_cfg.get("dropout", 0.0)
    enc_emb_dropout = enc_cfg["embeddings"].get("dropout", enc_dropout)
    if enc_cfg.get("type", "recurrent") == "transformer":
        assert enc_cfg["embeddings"]["embedding_dim"] == enc_cfg["hidden_size"], (
            "for transformer, emb_size must be "
            "the same as hidden_size")
        emb_size = src_embed.embedding_dim
        encoder = TransformerEncoder(
            **enc_cfg,
            emb_size=emb_size,
            emb_dropout=enc_emb_dropout,
            pad_index=src_pad_index,
        )
    else:
        encoder = RecurrentEncoder(
            **enc_cfg,
            emb_size=src_embed.embedding_dim,
            emb_dropout=enc_emb_dropout,
        )

    # build decoder
    dec_dropout = dec_cfg.get("dropout", 0.0)
    dec_emb_dropout = dec_cfg["embeddings"].get("dropout", dec_dropout)
    if dec_cfg.get("type", "transformer") == "transformer":
        decoder = TransformerDecoder(
            **dec_cfg,
            encoder=encoder,
            vocab_size=len(trg_vocab),
            emb_size=trg_embed.embedding_dim,
            emb_dropout=dec_emb_dropout,
        )
    else:
        decoder = RecurrentDecoder(
            **dec_cfg,
            encoder=encoder,
            vocab_size=len(trg_vocab),
            emb_size=trg_embed.embedding_dim,
            emb_dropout=dec_emb_dropout,
        )

    model = Model(
        encoder=encoder,
        decoder=decoder,
        src_embed=src_embed,
        trg_embed=trg_embed,
        src_vocab=src_vocab,
        trg_vocab=trg_vocab,
    )

    # tie softmax layer with trg embeddings
    if cfg.get("tied_softmax", False):
        if trg_embed.lut.weight.shape == model.decoder.output_layer.weight.shape:
            # (also) share trg embeddings and softmax layer:
            model.decoder.output_layer.weight = trg_embed.lut.weight
        else:
            raise ConfigurationError(
                "For tied_softmax, the decoder embedding_dim and decoder hidden_size "
                "must be the same. The decoder must be a Transformer.")

    # custom initialization of model parameters
    initialize_model(model, cfg, src_pad_index, trg_pad_index)

    # initialize embeddings from file
    enc_embed_path = enc_cfg["embeddings"].get("load_pretrained", None)
    dec_embed_path = dec_cfg["embeddings"].get("load_pretrained", None)
    if enc_embed_path and src_vocab is not None:
        logger.info("Loading pretrained src embeddings...")
        model.src_embed.load_from_file(Path(enc_embed_path), src_vocab)
    if (dec_embed_path and src_vocab is not None
            and not cfg.get("tied_embeddings", False)):
        logger.info("Loading pretrained trg embeddings...")
        model.trg_embed.load_from_file(Path(dec_embed_path), trg_vocab)

    logger.info("Enc-dec model built.")
    return model



from test.unit.test_helpers import TensorTestCase

import torch
from torch.nn import GRU, LSTM

from joeynmt.encoders import RecurrentEncoder


class TestRecurrentEncoder(TensorTestCase):

    def setUp(self):
        self.emb_size = 10
        self.num_layers = 3
        self.hidden_size = 7
        seed = 42
        torch.manual_seed(seed)

    def test_recurrent_encoder_size(self):
        for bidirectional in [True, False]:
            directional_factor = 2 if bidirectional else 1
            encoder = RecurrentEncoder(
                hidden_size=self.hidden_size,
                emb_size=self.emb_size,
                num_layers=self.num_layers,
                bidirectional=bidirectional,
            )
            self.assertEqual(encoder.rnn.hidden_size, self.hidden_size)
            # output size is affected by bidirectionality
            self.assertEqual(encoder.output_size, self.hidden_size * directional_factor)
            self.assertEqual(encoder.rnn.bidirectional, bidirectional)

    def test_recurrent_encoder_type(self):
        valid_rnn_types = {"gru": GRU, "lstm": LSTM}
        for name, obj in valid_rnn_types.items():
            encoder = RecurrentEncoder(rnn_type=name)
            self.assertEqual(type(encoder.rnn), obj)

    def test_recurrent_input_dropout(self):
        drop_prob = 0.5
        encoder = RecurrentEncoder(dropout=drop_prob, emb_dropout=drop_prob)
        input_tensor = torch.Tensor([2, 3, 1, -1])
        encoder.train()
        dropped = encoder.emb_dropout(input=input_tensor)
        # eval switches off dropout
        encoder.eval()
        no_drop = encoder.emb_dropout(input=input_tensor)
        # when dropout is applied, remaining values are divided by drop_prob
        self.assertGreaterEqual((no_drop - (drop_prob * dropped)).abs().sum(), 0)

        drop_prob = 1.0
        encoder = RecurrentEncoder(dropout=drop_prob, emb_dropout=drop_prob)
        all_dropped = encoder.emb_dropout(input=input_tensor)
        self.assertEqual(all_dropped.sum(), 0)
        encoder.eval()
        none_dropped = encoder.emb_dropout(input=input_tensor)
        self.assertTensorEqual(no_drop, none_dropped)
        self.assertTensorEqual((no_drop - all_dropped), no_drop)

    def test_recurrent_freeze(self):
        encoder = RecurrentEncoder(freeze=True)
        for _, p in encoder.named_parameters():
            self.assertFalse(p.requires_grad)

    def test_recurrent_forward(self):
        time_dim = 4
        batch_size = 2
        bidirectional = True
        directions = 2 if bidirectional else 1
        encoder = RecurrentEncoder(
            emb_size=self.emb_size,
            num_layers=self.num_layers,
            hidden_size=self.hidden_size,
            bidirectional=bidirectional,
        )
        x = torch.rand(size=(batch_size, time_dim, self.emb_size))
        # no padding, no mask
        x_length = torch.Tensor([time_dim] * batch_size).int()
        mask = torch.ones_like(x)
        output, hidden = encoder(embed_src=x, src_length=x_length, mask=mask)
        self.assertEqual(
            output.shape,
            torch.Size([batch_size, time_dim, directions * self.hidden_size]),
        )
        self.assertEqual(
            hidden.shape,
            torch.Size([batch_size, directions * self.hidden_size]),
        )
        hidden_target = torch.Tensor([
            [
                0.1323,
                0.0125,
                0.2900,
                -0.0725,
                -0.0102,
                -0.4405,
                0.1226,
                -0.3333,
                -0.3186,
                -0.2411,
                0.1790,
                0.1281,
                0.0739,
                -0.0536,
            ],
            [
                0.1431,
                0.0085,
                0.2828,
                -0.0933,
                -0.0139,
                -0.4525,
                0.0946,
                -0.3279,
                -0.3001,
                -0.2223,
                0.2023,
                0.0708,
                0.0131,
                -0.0124,
            ],
        ])
        output_target = torch.Tensor([[
            [
                [
                    0.0041,
                    0.0324,
                    0.0846,
                    -0.0056,
                    0.0353,
                    -0.2528,
                    0.0289,
                    -0.3333,
                    -0.3186,
                    -0.2411,
                    0.1790,
                    0.1281,
                    0.0739,
                    -0.0536,
                ],
                [
                    0.0159,
                    0.0248,
                    0.1496,
                    -0.0176,
                    0.0457,
                    -0.3839,
                    0.0780,
                    -0.3137,
                    -0.2731,
                    -0.2310,
                    0.1866,
                    0.0758,
                    0.0366,
                    -0.0069,
                ],
                [
                    0.0656,
                    0.0168,
                    0.2182,
                    -0.0391,
                    0.0214,
                    -0.4389,
                    0.1100,
                    -0.2625,
                    -0.1970,
                    -0.2249,
                    0.1374,
                    0.0337,
                    0.0139,
                    0.0284,
                ],
                [
                    0.1323,
                    0.0125,
                    0.2900,
                    -0.0725,
                    -0.0102,
                    -0.4405,
                    0.1226,
                    -0.1649,
                    -0.1023,
                    -0.1823,
                    0.0712,
                    0.0039,
                    -0.0228,
                    0.0444,
                ],
            ],
            [
                [
                    0.0296,
                    0.0254,
                    0.1007,
                    -0.0225,
                    0.0207,
                    -0.2612,
                    0.0061,
                    -0.3279,
                    -0.3001,
                    -0.2223,
                    0.2023,
                    0.0708,
                    0.0131,
                    -0.0124,
                ],
                [
                    0.0306,
                    0.0096,
                    0.1566,
                    -0.0386,
                    0.0387,
                    -0.3958,
                    0.0556,
                    -0.3034,
                    -0.2701,
                    -0.2165,
                    0.2061,
                    0.0364,
                    -0.0012,
                    0.0184,
                ],
                [
                    0.0842,
                    0.0075,
                    0.2181,
                    -0.0696,
                    0.0121,
                    -0.4389,
                    0.0874,
                    -0.2432,
                    -0.1979,
                    -0.2168,
                    0.1519,
                    0.0066,
                    -0.0080,
                    0.0485,
                ],
                [
                    0.1431,
                    0.0085,
                    0.2828,
                    -0.0933,
                    -0.0139,
                    -0.4525,
                    0.0946,
                    -0.1608,
                    -0.1140,
                    -0.1646,
                    0.0796,
                    -0.0202,
                    -0.0207,
                    0.0379,
                ],
            ],
        ]])
        self.assertTensorAlmostEqual(hidden_target, hidden)
        self.assertTensorAlmostEqual(output_target, output)



from test.unit.test_helpers import TensorTestCase

import torch
from torch.utils.data import BatchSampler, DataLoader, RandomSampler, SequentialSampler

from joeynmt.batch import Batch
from joeynmt.data import TokenBatchSampler, load_data, make_data_iter


class TestBatch(TensorTestCase):

    def setUp(self):
        # minimal data config
        data_cfg = {
            "train": "test/data/toy/train",
            "dev": "test/data/toy/dev",
            "src": {
                "lang": "de",
                "level": "char",
                "lowercase": True,
                "max_length": 20,
            },
            "trg": {
                "lang": "en",
                "level": "char",
                "lowercase": True,
                "max_length": 20,
            },
            "dataset_type": "plain",
        }

        # load the data
        (
            self.src_vocab,
            self.trg_vocab,
            self.train_data,
            self.dev_data,
            _,
        ) = load_data(data_cfg, datasets=["train", "dev"])
        self.pad_index = self.trg_vocab.pad_index
        # random seeds
        self.seed = 42

    def testBatchTrainIterator(self):

        batch_size = 4
        # load  all sents, filtering happens during batch construction
        self.assertEqual(len(self.train_data), 1000)

        # make data iterator
        train_iter = make_data_iter(
            dataset=self.train_data,
            batch_size=batch_size,
            batch_type="sentence",
            shuffle=True,
            seed=self.seed,
            pad_index=self.pad_index,
            device=torch.device("cpu"),
        )
        self.assertTrue(isinstance(train_iter, DataLoader))
        self.assertEqual(train_iter.batch_sampler.batch_size, batch_size)
        self.assertTrue(isinstance(train_iter.batch_sampler, BatchSampler))
        self.assertTrue(isinstance(train_iter.batch_sampler.sampler,
                                   RandomSampler))  # shuffle=True
        initial_seed = train_iter.batch_sampler.sampler.generator.initial_seed()
        self.assertEqual(initial_seed, self.seed)

        expected_src0 = torch.LongTensor(
            [[27, 7, 5, 14, 5, 4, 27, 5, 9, 30, 6, 12, 5, 9, 15, 6, 17, 5, 6, 24, 3],
             [19, 25, 11, 37, 24, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
             [28, 23, 5, 7, 23, 5, 6, 4, 12, 11, 6, 25, 5, 4, 8, 16, 13, 31, 6, 24, 3],
             [12, 11, 8, 4, 7, 8, 10, 4, 28, 11, 8, 8, 7, 5, 9, 10, 24, 3, 1, 1, 1]])
        expected_src0_len = torch.LongTensor([21, 6, 21, 18])
        expected_trg0 = torch.LongTensor(
            [[7, 4, 14, 8, 6, 4, 8, 23, 4, 17, 13, 7, 10, 21, 5, 24, 3],
             [8, 28, 7, 18, 24, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
             [6, 13, 7, 10, 28, 4, 18, 8, 16, 24, 3, 1, 1, 1, 1, 1, 1],
             [9, 6, 4, 15, 9, 15, 24, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1]])
        expected_trg0_len = torch.LongTensor([17, 6, 11, 8])

        total_samples = 0
        for b in train_iter:
            self.assertTrue(isinstance(b, Batch))
            if total_samples == 0:
                self.assertTensorEqual(b.src, expected_src0)
                self.assertTensorEqual(b.src_length, expected_src0_len)
                self.assertTensorEqual(b.trg, expected_trg0)
                self.assertTensorEqual(b.trg_length, expected_trg0_len)
            total_samples += b.nseqs
            self.assertLessEqual(b.nseqs, batch_size)
        self.assertEqual(total_samples, 27)

    def testTokenBatchTrainIterator(self):

        batch_size = 50  # num of tokens in one batch
        # load all sents here, filtering happends during batch construction
        self.assertEqual(len(self.train_data), 1000)

        # make data iterator
        train_iter = make_data_iter(
            dataset=self.train_data,
            batch_size=batch_size,
            batch_type="token",
            shuffle=True,
            seed=self.seed,
            pad_index=self.pad_index,
            device=torch.device("cpu"),
        )
        self.assertTrue(isinstance(train_iter, DataLoader))
        self.assertEqual(train_iter.batch_sampler.batch_size, batch_size)
        self.assertTrue(isinstance(train_iter.batch_sampler, TokenBatchSampler))
        self.assertTrue(isinstance(train_iter.batch_sampler.sampler,
                                   RandomSampler))  # shuffle=True
        initial_seed = train_iter.batch_sampler.sampler.generator.initial_seed()
        self.assertEqual(initial_seed, self.seed)

        expected_src0 = torch.LongTensor(
            [[27, 7, 5, 14, 5, 4, 27, 5, 9, 30, 6, 12, 5, 9, 15, 6, 17, 5, 6, 24, 3],
             [19, 25, 11, 37, 24, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
             [28, 23, 5, 7, 23, 5, 6, 4, 12, 11, 6, 25, 5, 4, 8, 16, 13, 31, 6, 24, 3]])
        expected_src0_len = torch.LongTensor([21, 6, 21])
        expected_trg0 = torch.LongTensor(
            [[7, 4, 14, 8, 6, 4, 8, 23, 4, 17, 13, 7, 10, 21, 5, 24, 3],
             [8, 28, 7, 18, 24, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
             [6, 13, 7, 10, 28, 4, 18, 8, 16, 24, 3, 1, 1, 1, 1, 1, 1]])
        expected_trg0_len = torch.LongTensor([17, 6, 11])

        total_tokens = 0
        for b in train_iter:
            self.assertTrue(isinstance(b, Batch))
            if total_tokens == 0:
                self.assertTensorEqual(b.src, expected_src0)
                self.assertTensorEqual(b.src_length, expected_src0_len)
                self.assertTensorEqual(b.trg, expected_trg0)
                self.assertTensorEqual(b.trg_length, expected_trg0_len)
            total_tokens += b.ntokens
        self.assertEqual(total_tokens, 387)

    def testBatchDevIterator(self):

        batch_size = 3
        self.assertEqual(len(self.dev_data), 20)

        # make data iterator
        dev_iter = make_data_iter(
            dataset=self.dev_data,
            batch_size=batch_size,
            batch_type="sentence",
            shuffle=False,
            pad_index=self.pad_index,
            device=torch.device("cpu"),
        )
        self.assertTrue(isinstance(dev_iter, DataLoader))
        self.assertEqual(dev_iter.batch_sampler.batch_size, batch_size)
        self.assertTrue(isinstance(dev_iter.batch_sampler, BatchSampler))
        self.assertTrue(isinstance(dev_iter.batch_sampler.sampler,
                                   SequentialSampler))  # shuffle=False

        expected_src0 = torch.LongTensor([[
            32, 11, 4, 22, 4, 11, 14, 8, 19, 4, 22, 4, 21, 11, 8, 4, 8, 19, 14, 14, 4,
            20, 7, 19, 13, 11, 16, 25, 7, 6, 17, 4, 8, 5, 7, 6, 4, 38, 3
        ],
                                          [
                                              7, 16, 13, 4, 23, 9, 5, 15, 5, 4, 18, 7,
                                              16, 13, 4, 22, 4, 12, 11, 8, 8, 4, 7, 16,
                                              13, 4, 12, 11, 4, 20, 7, 6, 4, 24, 3, 1,
                                              1, 1, 1
                                          ],
                                          [
                                              32, 11, 4, 22, 4, 17, 15, 10, 5, 6, 4, 10,
                                              11, 17, 4, 24, 3, 1, 1, 1, 1, 1, 1, 1, 1,
                                              1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
                                          ]])
        expected_src0_len = torch.LongTensor([39, 35, 17])
        expected_trg0 = torch.LongTensor([[
            18, 5, 11, 4, 26, 4, 11, 8, 4, 26, 4, 19, 13, 7, 6, 4, 9, 11, 4, 25, 9, 8,
            13, 7, 17, 28, 9, 10, 21, 4, 34, 3
        ],
                                          [
                                              9, 0, 20, 4, 13, 7, 22, 22, 18, 4, 6, 8,
                                              4, 25, 5, 4, 13, 5, 12, 5, 4, 24, 3, 1, 1,
                                              1, 1, 1, 1, 1, 1, 1
                                          ],
                                          [
                                              18, 5, 11, 4, 26, 4, 13, 5, 14, 14, 8, 4,
                                              24, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                                              1, 1, 1, 1, 1, 1
                                          ]])
        expected_trg0_len = torch.LongTensor([32, 23, 14])

        total_samples = 0
        for b in dev_iter:
            self.assertTrue(isinstance(b, Batch))

            # test the sorting by src length
            before_sort = b.src_length
            b.sort_by_src_length()
            after_sort = b.src_length
            self.assertTensorEqual(
                torch.sort(before_sort, descending=True)[0], after_sort)
            if total_samples == 0:
                self.assertTensorEqual(b.src, expected_src0)
                self.assertTensorEqual(b.src_length, expected_src0_len)
                self.assertTensorEqual(b.trg, expected_trg0)
                self.assertTensorEqual(b.trg_length, expected_trg0_len)
            total_samples += b.nseqs
            self.assertLessEqual(b.nseqs, batch_size)
        self.assertEqual(total_samples, len(self.dev_data))



# coding: utf-8
"""
Embedding module
"""

import logging
import math
from pathlib import Path
from typing import Dict

import torch
from torch import Tensor, nn

from joeynmt.helpers import freeze_params
from joeynmt.vocabulary import Vocabulary

logger = logging.getLogger(__name__)


class Embeddings(nn.Module):
    """
    Simple embeddings class
    """

    def __init__(
        self,
        embedding_dim: int = 64,
        scale: bool = False,
        vocab_size: int = 0,
        padding_idx: int = 1,
        freeze: bool = False,
        **kwargs,
    ):
        """
        Create new embeddings for the vocabulary.
        Use scaling for the Transformer.

        :param embedding_dim:
        :param scale:
        :param vocab_size:
        :param padding_idx:
        :param freeze: freeze the embeddings during training
        """
        # pylint: disable=unused-argument
        super().__init__()

        self.embedding_dim = embedding_dim
        self.scale = scale
        self.vocab_size = vocab_size
        self.lut = nn.Embedding(vocab_size, self.embedding_dim, padding_idx=padding_idx)

        if freeze:
            freeze_params(self)

    def forward(self, x: Tensor) -> Tensor:
        """
        Perform lookup for input `x` in the embedding table.

        :param x: index in the vocabulary
        :return: embedded representation for `x`
        """
        if self.scale:
            return self.lut(x) * math.sqrt(self.embedding_dim)
        return self.lut(x)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}("
                f"embedding_dim={self.embedding_dim}, "
                f"vocab_size={self.vocab_size})")

    # from fairseq
    def load_from_file(self, embed_path: Path, vocab: Vocabulary) -> None:
        """Load pretrained embedding weights from text file.

        - First line is expected to contain vocabulary size and dimension.
          The dimension has to match the model's specified embedding size,
          the vocabulary size is used in logging only.
        - Each line should contain word and embedding weights
          separated by spaces.
        - The pretrained vocabulary items that are not part of the
          joeynmt's vocabulary will be ignored (not loaded from the file).
        - The initialization (specified in config["model"]["embed_initializer"])
          of joeynmt's vocabulary items that are not part of the
          pretrained vocabulary will be kept (not overwritten in this func).
        - This function should be called after initialization!

        Example:
            2 5
            the -0.0230 -0.0264  0.0287  0.0171  0.1403
            at -0.0395 -0.1286  0.0275  0.0254 -0.0932

        :param embed_path: embedding weights text file
        :param vocab: Vocabulary object
        """
        # pylint: disable=logging-too-many-args

        embed_dict: Dict[int, Tensor] = {}
        # parse file
        with embed_path.open("r", encoding="utf-8", errors="ignore") as f_embed:
            vocab_size, d = map(int, f_embed.readline().split())
            assert self.embedding_dim == d, "Embedding dimension doesn't match."
            for line in f_embed.readlines():
                tokens = line.rstrip().split(" ")
                if tokens[0] in vocab.specials or not vocab.is_unk(tokens[0]):
                    embed_dict[vocab.lookup(tokens[0])] = torch.FloatTensor(
                        [float(t) for t in tokens[1:]])

            logger.warning(
                "Loaded %d of %d (%%) tokens in the pre-trained WE.",
                len(embed_dict),
                vocab_size,
                len(embed_dict) / vocab_size,
            )

        # assign
        for idx, weights in embed_dict.items():
            if idx < self.vocab_size:
                assert self.embedding_dim == len(weights)
                self.lut.weight.data[idx] = weights

        logger.warning(
            "Loaded %d of %d (%%) tokens of the JoeyNMT's vocabulary.",
            len(embed_dict),
            len(vocab),
            len(embed_dict) / len(vocab),
        )



import unittest

import torch

from joeynmt.data import load_data, make_data_iter


class TestData(unittest.TestCase):

    def setUp(self):
        self.train_path = "test/data/toy/train"
        self.dev_path = "test/data/toy/dev"
        self.test_path = "test/data/toy/test"
        self.max_length = 10
        self.seed = 42

        # minimal data config
        self.data_cfg = {
            "train": self.train_path,
            "dev": self.dev_path,
            "src": {
                "lang": "de",
                "level": "word",
                "lowercase": False,
                "max_length": self.max_length,
            },
            "trg": {
                "lang": "en",
                "level": "word",
                "lowercase": False,
                "max_length": self.max_length,
            },
            "dataset_type": "plain",
        }

    def testIteratorBatchType(self):

        current_cfg = self.data_cfg.copy()

        # load toy data
        _, trg_vocab, train_data, _, _ = load_data(current_cfg, datasets=["train"])

        # make batches by number of sentences
        train_iter = iter(
            make_data_iter(
                train_data,
                batch_size=10,
                batch_type="sentence",
                shuffle=True,
                seed=self.seed,
                pad_index=trg_vocab.pad_index,
                device=torch.device("cpu"),
                num_workers=0,
            ))
        batch = next(train_iter)

        self.assertEqual(batch.src.shape[0], 10)
        self.assertEqual(batch.trg.shape[0], 10)

        # make batches by number of tokens
        train_iter = iter(
            make_data_iter(
                train_data,
                batch_size=100,
                batch_type="token",
                shuffle=True,
                seed=self.seed,
                pad_index=trg_vocab.pad_index,
                device=torch.device("cpu"),
                num_workers=0,
            ))
        _ = next(train_iter)  # skip a batch
        _ = next(train_iter)  # skip another batch
        batch = next(train_iter)

        self.assertEqual(batch.src.shape, (9, 10))
        self.assertLessEqual(batch.ntokens, 77)



# coding: utf-8
"""
This modules holds methods for generating predictions from a model.
"""
import logging
import math
import sys
import time
from functools import partial
from itertools import zip_longest
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import torch
from torch.utils.data import Dataset

from joeynmt.data import load_data, make_data_iter
from joeynmt.datasets import build_dataset
from joeynmt.helpers import (
    expand_reverse_index,
    load_checkpoint,
    load_config,
    make_logger,
    parse_test_args,
    parse_train_args,
    resolve_ckpt_path,
    set_seed,
    store_attention_plots,
    write_list_to_file,
)
from joeynmt.metrics import bleu, chrf, sequence_accuracy, token_accuracy
from joeynmt.model import Model, _DataParallel, build_model
from joeynmt.search import search
from joeynmt.tokenizers import build_tokenizer
from joeynmt.vocabulary import build_vocab

logger = logging.getLogger(__name__)


def predict(
    model: Model,
    data: Dataset,
    device: torch.device,
    n_gpu: int,
    compute_loss: bool = False,
    normalization: str = "batch",
    num_workers: int = 0,
    cfg: Dict = None,
) -> Tuple[Dict[str, float], List[str], List[str], List[List[str]], List[np.ndarray],
           List[np.ndarray], ]:
    """
    Generates translations for the given data.
    If `compute_loss` is True and references are given, also computes the loss.

    :param model: model module
    :param data: dataset for validation
    :param device: torch device
    :param n_gpu: number of GPUs
    :param compute_loss: whether to computes a scalar loss for given inputs and targets
    :param normalization: one of {`batch`, `tokens`, `none`}
    :param num_workers: number of workers for `collate_fn()` in data iterator
    :param cfg: `testing` section in yaml config file
    :return:
        - valid_scores: (dict) current validation scores,
        - valid_ref: (list) validation references,
        - valid_hyp: (list) validation hypotheses,
        - decoded_valid: (list) token-level validation hypotheses (before post-process),
        - valid_sequence_scores: (list) log probabilities for validation hypotheses
        - valid_attention_scores: (list) attention scores for validation hypotheses
    """
    # pylint: disable=too-many-branches,too-many-statements
    # parse test cfg
    (
        eval_batch_size,
        eval_batch_type,
        max_output_length,
        min_output_length,
        eval_metrics,
        sacrebleu_cfg,
        beam_size,
        beam_alpha,
        n_best,
        return_attention,
        return_prob,
        generate_unk,
        repetition_penalty,
        no_repeat_ngram_size,
    ) = parse_test_args(cfg)

    if return_prob == "ref":  # no decoding needed
        decoding_description = ""
    else:
        decoding_description = (  # write the decoding strategy in the log
            " (Greedy decoding with " if beam_size < 2 else
            f" (Beam search with beam_size={beam_size}, beam_alpha={beam_alpha}, "
            f"n_best={n_best}, ")
        decoding_description += (
            f"min_output_length={min_output_length}, "
            f"max_output_length={max_output_length}, "
            f"return_prob='{return_prob}', generate_unk={generate_unk}, "
            f"repetition_penalty={repetition_penalty}, "
            f"no_repeat_ngram_size={no_repeat_ngram_size})")
    logger.info("Predicting %d example(s)...%s", len(data), decoding_description)

    assert eval_batch_size >= n_gpu, "`batch_size` must be bigger than `n_gpu`."
    # **CAUTION:** a batch will be expanded to batch.nseqs * beam_size, and it might
    # cause an out-of-memory error.
    # if batch_size > beam_size:
    #     batch_size //= beam_size

    valid_iter = make_data_iter(
        dataset=data,
        batch_size=eval_batch_size,
        batch_type=eval_batch_type,
        shuffle=False,
        num_workers=num_workers,
        pad_index=model.pad_index,
        device=device,
    )

    # disable dropout
    model.eval()

    # place holders for scores
    valid_scores = {"loss": float("nan"), "acc": float("nan"), "ppl": float("nan")}
    all_outputs = []
    valid_attention_scores = []
    valid_sequence_scores = []
    total_loss = 0
    total_nseqs = 0
    total_ntokens = 0
    total_n_correct = 0
    output, ref_scores, hyp_scores, attention_scores = None, None, None, None

    gen_start_time = time.time()
    for batch in valid_iter:
        total_nseqs += batch.nseqs  # number of sentences in the current batch

        # sort batch now by src length and keep track of order
        reverse_index = batch.sort_by_src_length()
        sort_reverse_index = expand_reverse_index(reverse_index, n_best)

        # run as during training to get validation loss (e.g. xent)
        if compute_loss and batch.has_trg:
            assert model.loss_function is not None

            # don't track gradients during validation
            with torch.no_grad():
                batch_loss, log_probs, _, n_correct = model(return_type="loss",
                                                            **vars(batch))
                # sum over multiple gpus
                batch_loss = batch.normalize(batch_loss, "sum", n_gpu=n_gpu)
                n_correct = batch.normalize(n_correct, "sum", n_gpu=n_gpu)
                if return_prob == "ref":
                    ref_scores = batch.score(log_probs)
                    output = batch.trg

            total_loss += batch_loss.item()  # cast Tensor to float
            total_n_correct += n_correct.item()  # cast Tensor to int
            total_ntokens += batch.ntokens

        # if return_prob == "ref", then no search needed.
        # (just look up the prob of the ground truth.)
        if return_prob != "ref":
            # run search as during inference to produce translations
            output, hyp_scores, attention_scores = search(
                model=model,
                batch=batch,
                beam_size=beam_size,
                beam_alpha=beam_alpha,
                max_output_length=max_output_length,
                n_best=n_best,
                return_attention=return_attention,
                return_prob=return_prob,
                generate_unk=generate_unk,
                repetition_penalty=repetition_penalty,
                no_repeat_ngram_size=no_repeat_ngram_size,
            )

        # sort outputs back to original order
        all_outputs.extend(output[sort_reverse_index])  # either hyp or ref
        valid_attention_scores.extend(attention_scores[sort_reverse_index]
                                      if attention_scores is not None else [])
        valid_sequence_scores.extend(
            ref_scores[sort_reverse_index] if ref_scores is not None
            and ref_scores.shape[0] == len(sort_reverse_index) else
            hyp_scores[sort_reverse_index] if hyp_scores is not None and hyp_scores.
            shape[0] == len(sort_reverse_index) else [])
    gen_duration = time.time() - gen_start_time

    assert total_nseqs == len(data), (total_nseqs, len(data))
    assert len(all_outputs) == len(data) * n_best, (len(all_outputs), len(data), n_best)

    if compute_loss:
        if normalization == "batch":
            normalizer = total_nseqs
        elif normalization == "tokens":
            normalizer = total_ntokens
        elif normalization == "none":
            normalizer = 1

        # avoid zero division
        assert normalizer > 0
        assert total_ntokens > 0

        # normalized loss
        valid_scores["loss"] = total_loss / normalizer
        # accuracy before decoding
        valid_scores["acc"] = total_n_correct / total_ntokens
        # exponent of token-level negative log likelihood
        valid_scores["ppl"] = math.exp(total_loss / total_ntokens)

    # decode ids back to str symbols (cut-off AFTER eos; eos itself is included.)
    decoded_valid = model.trg_vocab.arrays_to_sentences(arrays=all_outputs,
                                                        cut_at_eos=True)
    # TODO: `valid_sequence_scores` should have the same seq length as `decoded_valid`
    #     -> needed to be cut-off at eos synchronously

    if return_prob == "ref":  # no evaluation needed
        logger.info(
            "Evaluation result (scoring) %s, duration: %.4f[sec]",
            ", ".join([
                f"{eval_metric}: {valid_scores[eval_metric]:6.2f}"
                for eval_metric in ["loss", "ppl", "acc"]
            ]),
            gen_duration,
        )
        return valid_scores, None, None, decoded_valid, valid_sequence_scores, None

    # retrieve detokenized hypotheses and references
    valid_hyp = [
        data.tokenizer[data.trg_lang].post_process(s, generate_unk=generate_unk)
        for s in decoded_valid
    ]
    valid_ref = data.trg  # not length-filtered, not duplicated for n_best > 1

    # if references are given, evaluate 1best generation against them
    if data.has_trg:
        valid_hyp_1best = (valid_hyp if n_best == 1 else
                           [valid_hyp[i] for i in range(0, len(valid_hyp), n_best)])
        assert len(valid_hyp_1best) == len(valid_ref), (valid_hyp_1best, valid_ref)

        eval_start_time = time.time()

        # evaluate with metrics on full dataset
        for eval_metric in eval_metrics:
            if eval_metric == "bleu":
                valid_scores[eval_metric] = bleu(
                    valid_hyp_1best,
                    valid_ref,  # detokenized ref
                    **sacrebleu_cfg,
                )
            elif eval_metric == "chrf":
                valid_scores[eval_metric] = chrf(
                    valid_hyp_1best,
                    valid_ref,  # detokenized ref
                    **sacrebleu_cfg,
                )
            elif eval_metric == "token_accuracy":
                decoded_valid_1best = (decoded_valid if n_best == 1 else [
                    decoded_valid[i] for i in range(0, len(decoded_valid), n_best)
                ])
                valid_scores[eval_metric] = token_accuracy(
                    decoded_valid_1best,
                    data.get_list(lang=data.trg_lang, tokenized=True),  # tokenized ref
                )
            elif eval_metric == "sequence_accuracy":
                valid_scores[eval_metric] = sequence_accuracy(
                    valid_hyp_1best, valid_ref)

        eval_duration = time.time() - eval_start_time
        score_str = ", ".join([
            f"{eval_metric}: {valid_scores[eval_metric]:6.2f}"
            for eval_metric in eval_metrics + ["loss", "ppl", "acc"]
            if not math.isnan(valid_scores[eval_metric])
        ])
        logger.info(
            "Evaluation result (%s) %s, generation: %.4f[sec], evaluation: %.4f[sec]",
            "beam search" if beam_size > 1 else "greedy",
            score_str,
            gen_duration,
            eval_duration,
        )
    else:
        logger.info("Generation took %.4f[sec]. (No references given)", gen_duration)

    return (
        valid_scores,
        valid_ref,
        valid_hyp,
        decoded_valid,
        valid_sequence_scores,
        valid_attention_scores,
    )


def test(
    cfg_file,
    ckpt: str,
    output_path: str = None,
    datasets: dict = None,
    save_attention: bool = False,
    save_scores: bool = False,
) -> None:
    """
    Main test function. Handles loading a model from checkpoint, generating
    translations, storing them, and plotting attention.

    :param cfg_file: path to configuration file
    :param ckpt: path to checkpoint to load
    :param output_path: path to output
    :param datasets: datasets to predict
    :param save_attention: whether to save attention visualizations
    :param save_scores: whether to save scores
    """
    # pylint: disable=too-many-branches
    cfg = load_config(Path(cfg_file))
    # parse train cfg
    model_dir, load_model, device, n_gpu, num_workers, normalization = parse_train_args(
        cfg["training"], mode="prediction")

    if len(logger.handlers) == 0:
        _ = make_logger(model_dir, mode="test")  # version string returned

    # load the data
    if datasets is None:
        src_vocab, trg_vocab, _, dev_data, test_data = load_data(
            data_cfg=cfg["data"], datasets=["dev", "test"])
        data_to_predict = {"dev": dev_data, "test": test_data}
    else:  # avoid to load data again
        data_to_predict = {"dev": datasets["dev"], "test": datasets["test"]}
        src_vocab = datasets["src_vocab"]
        trg_vocab = datasets["trg_vocab"]

    # build model and load parameters into it
    model = build_model(cfg["model"], src_vocab=src_vocab, trg_vocab=trg_vocab)
    model.log_parameters_list()

    # check options
    if save_attention:
        if cfg["model"]["decoder"]["type"] == "transformer":
            assert cfg["testing"].get("beam_size", 1) == 1, (
                "Attention plots can be saved with greedy decoding only. Please set "
                "`beam_size: 1` in the config.")
        cfg["testing"]["save_attention"] = True
    return_prob = cfg["testing"].get("return_prob", "none")
    if save_scores:
        assert output_path, "Please specify --output_path for saving scores."
        if return_prob == "none":
            logger.warning("Please specify prob type: {`ref` or `hyp`} in the config. "
                           "Scores will not be saved.")
            save_scores = False
        elif return_prob == "ref":
            assert cfg["testing"].get("beam_size", 1) == 1, (
                "Scores of given references can be computed with greedy decoding only."
                "Please set `beam_size: 1` in the config.")
            model.loss_function = (  # need to instantiate loss func to compute scores
                cfg["training"].get("loss_type", "crossentropy"),
                cfg["training"].get("label_smoothing", 0.1),
            )

    # when checkpoint is not specified, take latest (best) from model dir
    ckpt = resolve_ckpt_path(ckpt, load_model, model_dir)

    # load model checkpoint
    model_checkpoint = load_checkpoint(ckpt, device=device)

    # restore model and optimizer parameters
    model.load_state_dict(model_checkpoint["model_state"])
    if device.type == "cuda":
        model.to(device)

    # multi-gpu eval
    if n_gpu > 1 and not isinstance(model, torch.nn.DataParallel):
        model = _DataParallel(model)

    # set the random seed
    set_seed(seed=cfg["training"].get("random_seed", 42))

    for data_set_name, data_set in data_to_predict.items():
        if data_set is None:
            continue

        logger.info(
            "%s on %s set...",
            "Scoring" if return_prob == "ref" else "Decoding",
            data_set_name,
        )
        _, _, hypotheses, hypotheses_raw, sequence_scores, attention_scores, = predict(
            model=model,
            data=data_set,
            compute_loss=save_scores,
            device=device,
            n_gpu=n_gpu,
            num_workers=num_workers,
            normalization=normalization,
            cfg=cfg["testing"],
        )

        if save_attention:
            if attention_scores:
                attention_file_name = f"{data_set_name}.{ckpt.stem}.att"
                attention_file_path = (model_dir / attention_file_name).as_posix()
                logger.info("Saving attention plots. This might take a while..")
                store_attention_plots(
                    attentions=attention_scores,
                    targets=hypotheses_raw,
                    sources=data_set.get_list(lang=data_set.src_lang, tokenized=True),
                    indices=range(len(hypotheses)),
                    output_prefix=attention_file_path,
                )
                logger.info("Attention plots saved to: %s", attention_file_path)
            else:
                logger.warning(
                    "Attention scores could not be saved. Note that attention scores "
                    "are not available when using beam search. Set beam_size to 1 for "
                    "greedy decoding.")

        if output_path is not None:
            if sequence_scores is not None and save_scores:
                # save scores
                output_path_scores = Path(f"{output_path}.{data_set_name}.scores")
                write_list_to_file(output_path_scores, sequence_scores)
                # save tokens
                output_path_tokens = Path(f"{output_path}.{data_set_name}.tokens")
                write_list_to_file(output_path_tokens, hypotheses_raw)
                logger.info(
                    "Scores and corresponding tokens saved to: %s.{scores|tokens}",
                    f"{output_path}.{data_set_name}",
                )
            if hypotheses is not None:
                output_path_set = Path(f"{output_path}.{data_set_name}")
                write_list_to_file(output_path_set, hypotheses)
                logger.info("Translations saved to: %s.", output_path_set)


def translate(
    cfg_file: str,
    ckpt: str = None,
    output_path: str = None,
) -> None:
    """
    Interactive translation function.
    Loads model from checkpoint and translates either the stdin input or asks for
    input to translate interactively. Translations and scores are printed to stdout.
    Note: The input sentences don't have to be pre-tokenized.

    :param cfg_file: path to configuration file
    :param ckpt: path to checkpoint to load
    :param output_path: path to output file
    """

    # pylint: disable=too-many-branches
    def _translate_data(test_data, cfg):
        """Translates given dataset, using parameters from outer scope."""
        _, _, hypotheses, trg_tokens, trg_scores, _ = predict(
            model=model,
            data=test_data,
            compute_loss=False,
            device=device,
            n_gpu=n_gpu,
            normalization="none",
            num_workers=num_workers,
            cfg=cfg,
        )
        return hypotheses, trg_tokens, trg_scores

    cfg = load_config(Path(cfg_file))
    # parse and validate cfg
    model_dir, load_model, device, n_gpu, num_workers, _ = parse_train_args(
        cfg["training"], mode="prediction")
    test_cfg = cfg["testing"]
    src_cfg = cfg["data"]["src"]
    trg_cfg = cfg["data"]["trg"]

    _ = make_logger(model_dir, mode="translate")
    # version string returned

    # when checkpoint is not specified, take latest (best) from model dir
    ckpt = resolve_ckpt_path(ckpt, load_model, model_dir)

    # read vocabs
    src_vocab, trg_vocab = build_vocab(cfg["data"], model_dir=model_dir)

    # build model and load parameters into it
    model = build_model(cfg["model"], src_vocab=src_vocab, trg_vocab=trg_vocab)

    # load model state from disk
    model_checkpoint = load_checkpoint(ckpt, device=device)
    model.load_state_dict(model_checkpoint["model_state"])

    if device.type == "cuda":
        model.to(device)

    tokenizer = build_tokenizer(cfg["data"])
    sequence_encoder = {
        src_cfg["lang"]: partial(src_vocab.sentences_to_ids, bos=False, eos=True),
        trg_cfg["lang"]: None,
    }
    test_data = build_dataset(
        dataset_type="stream",
        path=None,
        src_lang=src_cfg["lang"],
        trg_lang=trg_cfg["lang"],
        split="test",
        tokenizer=tokenizer,
        sequence_encoder=sequence_encoder,
    )

    # set the random seed
    set_seed(seed=cfg["training"].get("random_seed", 42))

    n_best = test_cfg.get("n_best", 1)
    beam_size = test_cfg.get("beam_size", 1)
    return_prob = test_cfg.get("return_prob", "none")
    if not sys.stdin.isatty():  # pylint: disable=too-many-nested-blocks
        # input stream given
        for line in sys.stdin.readlines():
            test_data.set_item(line.rstrip())
        all_hypotheses, tokens, scores = _translate_data(test_data, test_cfg)
        assert len(all_hypotheses) == len(test_data) * n_best

        if output_path is not None:
            # write to outputfile if given
            out_file = Path(output_path).expanduser()

            if n_best > 1:
                for n in range(n_best):
                    write_list_to_file(
                        out_file.parent / f"{out_file.stem}-{n}.{out_file.suffix}",
                        [
                            all_hypotheses[i]
                            for i in range(n, len(all_hypotheses), n_best)
                        ],
                    )
            else:
                write_list_to_file(out_file, all_hypotheses)

            logger.info("Translations saved to: %s.", out_file)

        else:
            # print to stdout
            for hyp in all_hypotheses:
                print(hyp)

    else:
        # enter interactive mode
        test_cfg["batch_size"] = 1  # CAUTION: this will raise an error if n_gpus > 1
        test_cfg["batch_type"] = "sentence"
        np.set_printoptions(linewidth=sys.maxsize)  # for printing scores in stdout
        while True:
            try:
                src_input = input("\nPlease enter a source sentence:\n")
                if not src_input.strip():
                    break

                # every line has to be made into dataset
                test_data.set_item(src_input.rstrip())
                hypotheses, tokens, scores = _translate_data(test_data, test_cfg)

                print("JoeyNMT:")
                for i, (hyp, token,
                        score) in enumerate(zip_longest(hypotheses, tokens, scores)):
                    assert hyp is not None, (i, hyp, token, score)
                    print(f"#{i + 1}: {hyp}")
                    if return_prob in ["hyp"]:
                        if beam_size > 1:  # beam search: sequence-level scores
                            print(f"\ttokens: {token}\n\tsequence score: {score[0]}")
                        else:  # greedy: token-level scores
                            assert len(token) == len(score), (token, score)
                            print(f"\ttokens: {token}\n\tscores: {score}")

                # reset cache
                test_data.cache = {}

            except (KeyboardInterrupt, EOFError):
                print("\nBye.")
                break



from test.unit.test_helpers import TensorTestCase

import torch
from torch.nn import GRU, LSTM

from joeynmt.decoders import RecurrentDecoder
from joeynmt.encoders import RecurrentEncoder


class TestRecurrentDecoder(TensorTestCase):

    def setUp(self):
        self.emb_size = 10
        self.num_layers = 3
        self.hidden_size = 6
        self.encoder_hidden_size = 3
        self.vocab_size = 5
        seed = 42
        torch.manual_seed(seed)

        bidi_encoder = RecurrentEncoder(
            emb_size=self.emb_size,
            num_layers=self.num_layers,
            hidden_size=self.encoder_hidden_size,
            bidirectional=True,
        )
        uni_encoder = RecurrentEncoder(
            emb_size=self.emb_size,
            num_layers=self.num_layers,
            hidden_size=self.encoder_hidden_size * 2,
            bidirectional=False,
        )
        self.encoders = [uni_encoder, bidi_encoder]

    def test_recurrent_decoder_size(self):
        # test all combinations of bridge, input_feeding, encoder directions
        for encoder in self.encoders:
            for init_hidden in ["bridge", "zero", "last"]:
                for input_feeding in [True, False]:
                    decoder = RecurrentDecoder(
                        hidden_size=self.hidden_size,
                        encoder=encoder,
                        attention="bahdanau",
                        emb_size=self.emb_size,
                        vocab_size=self.vocab_size,
                        num_layers=self.num_layers,
                        init_hidden=init_hidden,
                        input_feeding=input_feeding,
                    )
                    self.assertEqual(decoder.rnn.hidden_size, self.hidden_size)
                    self.assertEqual(decoder.att_vector_layer.out_features,
                                     self.hidden_size)
                    self.assertEqual(decoder.output_layer.out_features, self.vocab_size)
                    self.assertEqual(decoder.output_size, self.vocab_size)
                    self.assertEqual(decoder.rnn.bidirectional, False)

                    self.assertEqual(decoder.init_hidden_option, init_hidden)
                    if init_hidden == "bridge":
                        self.assertTrue(hasattr(decoder, "bridge_layer"))
                        self.assertEqual(decoder.bridge_layer.out_features,
                                         self.hidden_size)
                        self.assertEqual(
                            decoder.bridge_layer.in_features,
                            encoder.output_size,
                        )
                    else:
                        self.assertFalse(hasattr(decoder, "bridge_layer"))

                    if input_feeding:
                        self.assertEqual(
                            decoder.rnn_input_size,
                            self.emb_size + self.hidden_size,
                        )
                    else:
                        self.assertEqual(decoder.rnn_input_size, self.emb_size)

    def test_recurrent_decoder_type(self):
        valid_rnn_types = {"gru": GRU, "lstm": LSTM}
        for name, obj in valid_rnn_types.items():
            decoder = RecurrentDecoder(
                rnn_type=name,
                hidden_size=self.hidden_size,
                encoder=self.encoders[0],
                attention="bahdanau",
                emb_size=self.emb_size,
                vocab_size=self.vocab_size,
                num_layers=self.num_layers,
                init_hidden="zero",
                input_feeding=False,
            )
            self.assertEqual(type(decoder.rnn), obj)

    def test_recurrent_input_dropout(self):
        drop_prob = 0.5
        decoder = RecurrentDecoder(
            hidden_size=self.hidden_size,
            encoder=self.encoders[0],
            attention="bahdanau",
            emb_size=self.emb_size,
            vocab_size=self.vocab_size,
            num_layers=self.num_layers,
            init_hidden="zero",
            input_feeding=False,
            dropout=drop_prob,
            emb_dropout=drop_prob,
        )
        input_tensor = torch.Tensor([2, 3, 1, -1])
        decoder.train()
        dropped = decoder.emb_dropout(input=input_tensor)
        # eval switches off dropout
        decoder.eval()
        no_drop = decoder.emb_dropout(input=input_tensor)
        # when dropout is applied, remaining values are divided by drop_prob
        self.assertGreaterEqual((no_drop - (drop_prob * dropped)).abs().sum(), 0)

        drop_prob = 1.0
        decoder = RecurrentDecoder(
            hidden_size=self.hidden_size,
            encoder=self.encoders[0],
            attention="bahdanau",
            emb_size=self.emb_size,
            vocab_size=self.vocab_size,
            num_layers=self.num_layers,
            init_hidden="zero",
            input_feeding=False,
            dropout=drop_prob,
            emb_dropout=drop_prob,
        )
        all_dropped = decoder.emb_dropout(input=input_tensor)
        self.assertEqual(all_dropped.sum(), 0)
        decoder.eval()
        none_dropped = decoder.emb_dropout(input=input_tensor)
        self.assertTensorEqual(no_drop, none_dropped)
        self.assertTensorEqual((no_drop - all_dropped), no_drop)

    def test_recurrent_hidden_dropout(self):
        hidden_drop_prob = 0.5
        decoder = RecurrentDecoder(
            hidden_size=self.hidden_size,
            encoder=self.encoders[0],
            attention="bahdanau",
            emb_size=self.emb_size,
            vocab_size=self.vocab_size,
            num_layers=self.num_layers,
            init_hidden="zero",
            input_feeding=False,
            hidden_dropout=hidden_drop_prob,
        )
        input_tensor = torch.Tensor([2, 3, 1, -1])
        decoder.train()
        dropped = decoder.hidden_dropout(input=input_tensor)
        # eval switches off dropout
        decoder.eval()
        no_drop = decoder.hidden_dropout(input=input_tensor)
        # when dropout is applied, remaining values are divided by drop_prob
        self.assertGreaterEqual((no_drop - (hidden_drop_prob * dropped)).abs().sum(), 0)

        hidden_drop_prob = 1.0
        decoder = RecurrentDecoder(
            hidden_size=self.hidden_size,
            encoder=self.encoders[0],
            attention="bahdanau",
            emb_size=self.emb_size,
            vocab_size=self.vocab_size,
            num_layers=self.num_layers,
            init_hidden="zero",
            input_feeding=False,
            hidden_dropout=hidden_drop_prob,
        )
        all_dropped = decoder.hidden_dropout(input=input_tensor)
        self.assertEqual(all_dropped.sum(), 0)
        decoder.eval()
        none_dropped = decoder.hidden_dropout(input=input_tensor)
        self.assertTensorEqual(no_drop, none_dropped)
        self.assertTensorEqual((no_drop - all_dropped), no_drop)

    def test_recurrent_freeze(self):
        decoder = RecurrentDecoder(
            hidden_size=self.hidden_size,
            encoder=self.encoders[0],
            attention="bahdanau",
            emb_size=self.emb_size,
            vocab_size=self.vocab_size,
            num_layers=self.num_layers,
            init_hidden="zero",
            input_feeding=False,
            freeze=True,
        )
        for _, p in decoder.named_parameters():
            self.assertFalse(p.requires_grad)

    def test_recurrent_forward(self):
        time_dim = 4
        batch_size = 2
        # make sure the outputs match the targets
        decoder = RecurrentDecoder(
            hidden_size=self.hidden_size,
            encoder=self.encoders[0],
            attention="bahdanau",
            emb_size=self.emb_size,
            vocab_size=self.vocab_size,
            num_layers=self.num_layers,
            init_hidden="zero",
            input_feeding=False,
        )
        encoder_states = torch.rand(size=(batch_size, time_dim,
                                          self.encoders[0].output_size))
        trg_inputs = torch.ones(size=(batch_size, time_dim, self.emb_size))
        # no padding, no mask
        # x_length = torch.Tensor([time_dim]*batch_size).int()
        mask = torch.ones(size=(batch_size, 1, time_dim)).byte()
        output, hidden, att_probs, att_vectors = decoder(
            trg_inputs,
            encoder_hidden=encoder_states[:, -1, :],
            encoder_output=encoder_states,
            src_mask=mask,
            unroll_steps=time_dim,
            hidden=None,
            prev_att_vector=None,
        )
        self.assertEqual(output.shape,
                         torch.Size([batch_size, time_dim, self.vocab_size]))
        self.assertEqual(
            hidden.shape,
            torch.Size([batch_size, self.num_layers, self.hidden_size]),
        )
        self.assertEqual(att_probs.shape, torch.Size([batch_size, time_dim, time_dim]))
        self.assertEqual(
            att_vectors.shape,
            torch.Size([batch_size, time_dim, self.hidden_size]),
        )
        hidden_target = torch.Tensor([
            [
                [0.1814, 0.5468, -0.4717, -0.7580, 0.5834, -0.4018],
                [0.4649, 0.5484, -0.2702, 0.4545, 0.1983, 0.2771],
                [-0.1752, -0.4215, 0.1941, -0.3975, -0.2317, -0.5566],
            ],
            [
                [0.1814, 0.5468, -0.4717, -0.7580, 0.5834, -0.4018],
                [0.4649, 0.5484, -0.2702, 0.4545, 0.1983, 0.2771],
                [-0.1752, -0.4215, 0.1941, -0.3975, -0.2317, -0.5566],
            ],
        ])
        output_target = torch.Tensor([
            [
                [0.2702, -0.1988, -0.1985, -0.2998, -0.2564],
                [0.2719, -0.2075, -0.2017, -0.2988, -0.2595],
                [0.2720, -0.2143, -0.2084, -0.3024, -0.2537],
                [0.2714, -0.2183, -0.2135, -0.3061, -0.2468],
            ],
            [
                [0.2757, -0.1744, -0.1888, -0.3038, -0.2466],
                [0.2782, -0.1837, -0.1928, -0.3028, -0.2505],
                [0.2785, -0.1904, -0.1994, -0.3066, -0.2448],
                [0.2777, -0.1943, -0.2042, -0.3105, -0.2379],
            ],
        ])
        att_vectors_target = torch.Tensor([
            [
                [-0.6196, -0.0505, 0.4900, 0.6286, -0.5007, -0.3721],
                [-0.6389, -0.0337, 0.4998, 0.6458, -0.5052, -0.3579],
                [-0.6396, -0.0158, 0.5058, 0.6609, -0.5035, -0.3660],
                [-0.6348, -0.0017, 0.5090, 0.6719, -0.5013, -0.3771],
            ],
            [
                [-0.5697, -0.0887, 0.4515, 0.6128, -0.4713, -0.4068],
                [-0.5910, -0.0721, 0.4617, 0.6305, -0.4760, -0.3930],
                [-0.5918, -0.0544, 0.4680, 0.6461, -0.4741, -0.4008],
                [-0.5866, -0.0405, 0.4712, 0.6574, -0.4718, -0.4116],
            ],
        ])
        self.assertTensorAlmostEqual(hidden_target, hidden)
        self.assertTensorAlmostEqual(output_target, output)
        self.assertTensorAlmostEqual(att_vectors, att_vectors_target)
        # att_probs should be a distribution over the output vocabulary
        self.assertTensorAlmostEqual(att_probs.sum(2), torch.ones(batch_size, time_dim))



import tempfile
import unittest
from pathlib import Path

from joeynmt.data import load_data
from joeynmt.datasets import PlaintextDataset, TsvDataset
from joeynmt.helpers import read_list_from_file, write_list_to_file


class TestPlaintextDataset(unittest.TestCase):

    def setUp(self):
        self.train_path = "test/data/toy/train"
        self.dev_path = "test/data/toy/dev"
        self.test_path = "test/data/toy/test"
        self.max_length = 10
        self.min_length = 5
        self.seed = 42

        # minimal data config
        self.data_cfg = {
            "train": self.train_path,
            "dev": self.dev_path,
            "test": self.test_path,
            "src": {
                "lang": "de",
                "level": "word",
                "lowercase": False,
                "max_length": self.max_length,
                "min_length": self.min_length,
            },
            "trg": {
                "lang": "en",
                "level": "word",
                "lowercase": False,
                "max_length": self.max_length,
                "min_length": self.min_length,
            },
            "random_train_subset": 100,
            "random_dev_subset": 100,
            "dataset_type": "plain",
        }

    def testDataLoading(self):
        # load the data
        datasets = ["train", "dev"]
        for test_path in [None, self.test_path]:
            current_cfg = self.data_cfg.copy()
            if test_path is not None:
                if "test" not in datasets:
                    datasets.append("test")
                current_cfg["test"] = test_path
            else:
                if "test" in datasets:
                    datasets.remove("test")

            _, _, train_data, dev_data, test_data = load_data(current_cfg,
                                                              datasets=datasets)

            self.assertIs(type(train_data), PlaintextDataset)
            self.assertIs(type(dev_data), PlaintextDataset)

            self.assertEqual(train_data.split, "train")
            self.assertEqual(dev_data.split, "dev")

            self.assertTrue(train_data.has_trg)
            self.assertTrue(dev_data.has_trg)

            if test_path is None:
                self.assertIsNone(test_data)
            else:
                self.assertIs(type(test_data), PlaintextDataset)
                self.assertEqual(test_data.split, "test")
                self.assertFalse(test_data.has_trg)

            self.assertEqual(train_data.tokenizer[train_data.src_lang].max_length,
                             self.max_length)
            self.assertEqual(train_data.tokenizer[train_data.trg_lang].max_length,
                             self.max_length)

            self.assertEqual(train_data.tokenizer[train_data.src_lang].min_length,
                             self.min_length)
            self.assertEqual(train_data.tokenizer[train_data.trg_lang].min_length,
                             self.min_length)

            # check the number of examples loaded
            # NOTE: since tokenization is applied in batch construction,
            # we cannot compute the length and therefore cannot filter
            # examples out based on the length before batch iteration.
            expected_train_len, after_filtering = 1000, 290
            expected_testdev_len = 20  # dev and test have the same len
            self.assertEqual(len(train_data), expected_train_len)
            self.assertEqual(len(dev_data), expected_testdev_len)
            if test_path is None:
                self.assertIsNone(test_data)
            else:
                self.assertEqual(len(test_data), expected_testdev_len)

            train_ex = [train_data[i] for i in range(len(train_data))]
            train_ex = [(s, t) for s, t in train_ex if s is not None and t is not None]
            self.assertEqual(len(train_ex), after_filtering)

            # check the segmentation: src and trg attributes are lists
            train_src, train_trg = train_ex[0]
            dev_src, dev_trg = dev_data[0]
            self.assertIs(type(train_src), list)
            self.assertIs(type(train_trg), list)
            self.assertIs(type(dev_src), list)
            self.assertIs(type(dev_trg), list)
            if test_path is not None:
                test_src, test_trg = test_data[0]
                self.assertIs(type(test_src), list)
                self.assertIs(test_trg, None)

            # check the length filtering of the training examples
            src_len, trg_len = zip(*train_ex)
            self.assertTrue(
                all(self.min_length <= len(s) <= self.max_length for s in src_len))
            self.assertTrue(
                all(self.min_length <= len(t) <= self.max_length for t in trg_len))

            # check the lowercasing
            if current_cfg["src"]["lowercase"]:
                self.assertTrue(all(ex.lower() == ex for ex in train_data.src))
                self.assertTrue(all(ex.lower() == ex for ex in dev_data.src))
                if test_path is not None:
                    self.assertTrue(all(ex.lower() == ex for ex in test_data.src))
            if current_cfg["trg"]["lowercase"]:
                self.assertTrue(all(ex.lower() == ex for ex in train_data.trg))
                self.assertTrue(all(ex.lower() == ex for ex in dev_data.trg))

            # check dev: no length filtering
            dev_ex = [dev_data[i] for i in range(len(dev_data))]
            dev_src, dev_trg = zip(*dev_ex)
            self.assertEqual(len(dev_ex), expected_testdev_len)
            self.assertEqual(min([len(t) for t in dev_trg]), 4)
            self.assertEqual(max([len(t) for t in dev_trg]), 46)
            self.assertTrue(all(t is not None for t in dev_src))
            self.assertTrue(all(t is not None for t in dev_trg))

    def testRandomSubset(self):
        # Load data
        _, _, train_data, _, test_data = load_data(self.data_cfg,
                                                   datasets=["train", "test"])
        self.assertEqual(len(train_data), 1000)
        self.assertEqual(train_data.random_subset, 100)
        train_data.sample_random_subset(seed=self.seed)
        self.assertEqual(len(train_data), 100)

        train_data.reset_random_subset()
        self.assertEqual(len(train_data), 1000)

        # a random subset can be selected only when len(train_data) > n
        train_data.random_subset = 2000
        with self.assertRaises(AssertionError) as e:
            train_data.sample_random_subset(seed=self.seed)
        self.assertEqual("Can only subsample from train or dev set larger than 2000.",
                         str(e.exception))

        # a random subset should be selected for training only
        self.assertEqual(test_data.random_subset, -1)
        test_data.random_subset = 100
        with self.assertRaises(AssertionError) as e:
            test_data.sample_random_subset(seed=self.seed)
        self.assertEqual("Can only subsample from train or dev set larger than 100.",
                         str(e.exception))


class TestTsvDataset(unittest.TestCase):

    def setUp(self):
        # save toy data temporarily in tsv format
        def _read_write_sents(path, src_lang, trg_lang):
            with tempfile.NamedTemporaryFile(prefix="joeynmt_unittest_",
                                             suffix=".tsv",
                                             delete=False) as temp:
                tsv_file = Path(temp.name)
                src = read_list_from_file(Path(path).with_suffix(f".{src_lang}"))
                if trg_lang:
                    trg = read_list_from_file(Path(path).with_suffix(f".{trg_lang}"))
                    lines = [f"{src_lang}\t{trg_lang}"
                             ] + [f"{s}\t{t}" for s, t in zip(src, trg)]
                else:
                    lines = [src_lang] + src
                write_list_to_file(tsv_file, lines)
            return tsv_file

        train_file = _read_write_sents("test/data/toy/train", "de", "en")
        dev_file = _read_write_sents("test/data/toy/dev", "de", "en")
        test_file = _read_write_sents("test/data/toy/test", "de", None)

        self.max_length = 10
        self.min_length = 5

        # random seed for subsampling
        self.seed = 42

        # minimal data config
        self.data_cfg = {
            "train": (train_file.parent / train_file.stem).as_posix(),
            "dev": (dev_file.parent / dev_file.stem).as_posix(),
            "test": (test_file.parent / test_file.stem).as_posix(),
            "src": {
                "lang": "de",
                "level": "word",
                "lowercase": False,
                "max_length": self.max_length,
                "min_length": self.min_length,
            },
            "trg": {
                "lang": "en",
                "level": "word",
                "lowercase": False,
                "max_length": self.max_length,
                "min_length": self.min_length,
            },
            "random_train_subset": 100,
            "random_dev_subset": 100,
            "dataset_type": "tsv",
        }

    def tearDown(self):
        # delete tmp files
        train_file = Path(self.data_cfg["train"]).with_suffix(".tsv")
        dev_file = Path(self.data_cfg["dev"]).with_suffix(".tsv")
        test_file = Path(self.data_cfg["test"]).with_suffix(".tsv")

        if train_file.is_file():
            train_file.unlink()
        if dev_file.is_file():
            dev_file.unlink()
        if test_file.is_file():
            test_file.unlink()

    def testDataLoading(self):
        try:
            expected_train_len, after_filtering = 1000, 290
            expected_dev_len = 20
            _, _, train_data, dev_data, _ = load_data(self.data_cfg,
                                                      datasets=["train", "dev"])

            self.assertIs(type(train_data), TsvDataset)
            self.assertIs(type(dev_data), TsvDataset)

            self.assertEqual(len(train_data), expected_train_len)
            self.assertEqual(len(dev_data), expected_dev_len)

            self.assertEqual(train_data.split, "train")
            self.assertEqual(dev_data.split, "dev")

            self.assertTrue(train_data.has_trg)
            self.assertTrue(dev_data.has_trg)

            self.assertEqual(train_data.tokenizer[train_data.src_lang].max_length,
                             self.max_length)
            self.assertEqual(train_data.tokenizer[train_data.trg_lang].max_length,
                             self.max_length)

            self.assertEqual(train_data.tokenizer[train_data.src_lang].min_length,
                             self.min_length)
            self.assertEqual(train_data.tokenizer[train_data.trg_lang].min_length,
                             self.min_length)

            train_ex = [train_data[i] for i in range(len(train_data))]
            train_ex = [(s, t) for s, t in train_ex if s is not None and t is not None]
            self.assertEqual(len(train_ex), after_filtering)

            # check the length filtering of the training examples
            src_len, trg_len = zip(*train_ex)
            self.assertTrue(
                all(self.min_length <= len(s) <= self.max_length for s in src_len))
            self.assertTrue(
                all(self.min_length <= len(t) <= self.max_length for t in trg_len))

            dev_ex = [dev_data[i] for i in range(len(dev_data))]
            dev_src, dev_trg = zip(*dev_ex)
            self.assertEqual(len(dev_ex), expected_dev_len)
            self.assertEqual(min([len(t) for t in dev_trg]), 4)
            self.assertEqual(max([len(t) for t in dev_trg]), 46)
            self.assertTrue(all(t is not None for t in dev_src))
            self.assertTrue(all(t is not None for t in dev_trg))

        except ImportError as e:
            # need pandas installed.
            raise unittest.SkipTest(f"{e} Skip.")

    def testRandomSubset(self):
        try:
            # load the data
            _, _, train_data, _, test_data = load_data(self.data_cfg,
                                                       datasets=["train", "test"])
            self.assertEqual(len(train_data), 1000)
            self.assertEqual(train_data.random_subset, 100)
            train_data.sample_random_subset(seed=self.seed)
            self.assertEqual(len(train_data), 100)

            # a random subset can be selected only when len(train_data) > n
            train_data.random_subset = 2000
            with self.assertRaises(AssertionError) as e:
                train_data.sample_random_subset(seed=self.seed)
            self.assertEqual(
                "Can only subsample from train or dev set larger than 2000.",
                str(e.exception))

            # a random subset should be selected for training only
            self.assertEqual(test_data.random_subset, -1)
            test_data.random_subset = 100
            with self.assertRaises(AssertionError) as e:
                test_data.sample_random_subset(seed=self.seed)
            self.assertEqual(
                "Can only subsample from train or dev set larger than 100.",
                str(e.exception))

        except ImportError as e:
            raise unittest.SkipTest(f"{e} Skip.")



import unittest
from test.unit.test_helpers import TensorTestCase

from joeynmt.metrics import bleu, chrf, token_accuracy


class TestMetrics(TensorTestCase):

    def test_chrf_without_whitespace(self):
        hyp1 = ["t est"]
        ref1 = ["tez t"]
        score1 = chrf(hyp1, ref1, whitespace=False)
        hyp2 = ["test"]
        ref2 = ["tezt"]
        score2 = chrf(hyp2, ref2, whitespace=False)
        self.assertAlmostEqual(score1, score2)
        self.assertAlmostEqual(score1, 0.271, places=3)

    def test_chrf_with_whitespace(self):
        hyp = ["これはテストです。"]
        ref = ["これは テストです。"]
        score = chrf(hyp, ref, whitespace=True)
        self.assertAlmostEqual(score, 0.558, places=3)

    def test_bleu_13a(self):
        hyp = ["This is a test."]
        ref = ["this is a Tezt."]
        score = bleu(hyp, ref, tokenize="13a", lowercase=True)
        self.assertAlmostEqual(score, 42.729, places=3)

    def test_bleu_ja_mecab(self):
        try:
            hyp = ["これはテストです。"]
            ref = ["あれがテストです。"]
            score = bleu(hyp, ref, tokenize="ja-mecab")
            self.assertAlmostEqual(score, 39.764, places=3)
        except (ImportError, RuntimeError) as e:
            raise unittest.SkipTest(f"{e} Skip.")

    def test_token_acc_level_char(self):
        # if len(hyp) > len(ref)
        hyp = [list("tests")]
        ref = [list("tezt")]
        # level = "char"
        score = token_accuracy(hyp, ref)
        self.assertEqual(score, 60.0)

        # if len(hyp) < len(ref)
        hyp = [list("test")]
        ref = [list("tezts")]
        # level = "char"
        score = token_accuracy(hyp, ref)
        self.assertEqual(score, 75.0)



import copy
from test.unit.test_helpers import TensorTestCase

import torch
from torch import nn

from joeynmt.model import build_model
from joeynmt.vocabulary import Vocabulary


class TestModelInit(TensorTestCase):

    def setUp(self):
        self.seed = 42
        vocab_size = 30
        tokens = [f"tok{i:02d}" for i in range(vocab_size)]
        self.vocab = Vocabulary(tokens=tokens)
        self.hidden_size = 64

        self.cfg = {
            "model": {
                "tied_embeddings": False,
                "tied_softmax": False,
                "encoder": {
                    "type": "transformer",
                    "hidden_size": self.hidden_size,
                    "embeddings": {
                        "embedding_dim": self.hidden_size
                    },
                    "num_layers": 1,
                    "layer_norm": "pre",
                },
                "decoder": {
                    "type": "transformer",
                    "hidden_size": self.hidden_size,
                    "embeddings": {
                        "embedding_dim": self.hidden_size
                    },
                    "num_layers": 1,
                    "layer_norm": "pre",
                },
            }
        }

    def test_transformer_layer_norm_init(self):
        torch.manual_seed(self.seed)
        cfg = copy.deepcopy(self.cfg)

        src_vocab = trg_vocab = self.vocab

        model = build_model(cfg["model"], src_vocab=src_vocab, trg_vocab=trg_vocab)

        def check_layer_norm(m: nn.Module):
            for _, child in m.named_children():
                if isinstance(child, nn.LayerNorm):
                    self.assertTensorEqual(child.weight, torch.ones([self.hidden_size]))
                    self.assertTensorEqual(child.bias, torch.zeros([self.hidden_size]))
                else:
                    check_layer_norm(child)

        check_layer_norm(model)

    def test_transformer_layers_init(self):
        torch.manual_seed(self.seed)
        cfg = copy.deepcopy(self.cfg)
        cfg["model"]["initializer"] = "xavier_normal"
        cfg["model"]["encoder"]["num_layers"] = 6
        cfg["model"]["decoder"]["num_layers"] = 6

        src_vocab = trg_vocab = self.vocab

        model = build_model(cfg["model"], src_vocab=src_vocab, trg_vocab=trg_vocab)

        self.assertEqual(len(model.encoder.layers), 6)
        self.assertEqual(len(model.decoder.layers), 6)

        for layer in model.encoder.layers:
            self.assertEqual(layer.alpha, 1.417938140685523)
        for layer in model.decoder.layers:
            self.assertEqual(layer.alpha, 2.0597671439071177)

        self.assertTensorAlmostEqual(
            model.encoder.layers[0].src_src_att.q_layer.weight[:5, 0].data,
            torch.Tensor([-0.2093, -0.1066, -0.1455, -0.1146, 0.0760]),
        )
        self.assertTensorAlmostEqual(
            model.decoder.layers[0].src_trg_att.q_layer.weight[:5, 0].data,
            torch.Tensor([0.0072, -0.0241, 0.2873, -0.0417, -0.2752]),
        )
        self.assertTensorAlmostEqual(
            model.decoder.layers[0].trg_trg_att.q_layer.weight[:5, 0].data,
            torch.Tensor([-0.2140, 0.0942, 0.0203, 0.0417, 0.2482]),
        )



# coding: utf-8
"""
Module to implement training loss
"""
import torch
from torch import Tensor, nn
from torch.autograd import Variable
from torch.nn.modules.loss import _Loss


class XentLoss(nn.Module):
    """
    Cross-Entropy Loss with optional label smoothing
    """

    def __init__(self, pad_index: int, smoothing: float = 0.0):
        super().__init__()
        self.smoothing = smoothing
        self.pad_index = pad_index
        self.criterion: _Loss  # (type annotation)
        if self.smoothing <= 0.0:
            # standard xent loss
            self.criterion = nn.NLLLoss(ignore_index=self.pad_index, reduction="sum")
        else:
            # custom label-smoothed loss, computed with KL divergence loss
            self.criterion = nn.KLDivLoss(reduction="sum")

    def _smooth_targets(self, targets: Tensor, vocab_size: int) -> Variable:
        """
        Smooth target distribution. All non-reference words get uniform
        probability mass according to "smoothing".

        :param targets: target indices, batch*seq_len
        :param vocab_size: size of the output vocabulary
        :return: smoothed target distributions, batch*seq_len x vocab_size
        """
        # batch*seq_len x vocab_size
        smooth_dist = targets.new_zeros((targets.size(0), vocab_size)).float()
        # fill distribution uniformly with smoothing
        smooth_dist.fill_(self.smoothing / (vocab_size - 2))
        # assign true label the probability of 1-smoothing ("confidence")
        smooth_dist.scatter_(1, targets.unsqueeze(1).data, 1.0 - self.smoothing)
        # give padding probability of 0 everywhere
        smooth_dist[:, self.pad_index] = 0
        # masking out padding area (sum of probabilities for padding area = 0)
        padding_positions = torch.nonzero(targets.data == self.pad_index,
                                          as_tuple=False)
        if len(padding_positions) > 0:
            smooth_dist.index_fill_(0, padding_positions.squeeze(), 0.0)
        return Variable(smooth_dist, requires_grad=False)

    def _reshape(self, log_probs: Tensor, targets: Tensor) -> Tensor:
        vocab_size = log_probs.size(-1)

        # reshape log_probs to (batch*seq_len x vocab_size)
        log_probs_flat = log_probs.contiguous().view(-1, vocab_size)

        if self.smoothing > 0:
            targets_flat = self._smooth_targets(targets=targets.contiguous().view(-1),
                                                vocab_size=vocab_size)
            # targets: distributions with batch*seq_len x vocab_size
            assert log_probs_flat.size() == targets_flat.size(), (
                log_probs.size(),
                targets_flat.size(),
            )
        else:
            # targets: indices with batch*seq_len
            targets_flat = targets.contiguous().view(-1)
            assert log_probs_flat.size(0) == targets_flat.size(0), (
                log_probs.size(0),
                targets_flat.size(0),
            )

        return log_probs_flat, targets_flat

    def forward(self, log_probs: Tensor, **kwargs) -> Tensor:
        """
        Compute the cross-entropy between logits and targets.

        If label smoothing is used, target distributions are not one-hot, but
        "1-smoothing" for the correct target token and the rest of the
        probability mass is uniformly spread across the other tokens.

        :param log_probs: log probabilities as predicted by model
        :return: logits
        """
        assert "trg" in kwargs
        log_probs, targets = self._reshape(log_probs, kwargs["trg"])

        # compute loss
        logits = self.criterion(log_probs, targets)
        return logits

    def __repr__(self):
        return (f"{self.__class__.__name__}(criterion={self.criterion}, "
                f"smoothing={self.smoothing})")



# coding: utf-8
import unittest
from pathlib import Path

import numpy as np

from joeynmt.data import load_data
from joeynmt.helpers import read_list_from_file
from joeynmt.vocabulary import Vocabulary


class TestVocabulary(unittest.TestCase):

    def setUp(self):
        self.voc_limit = 1000
        self.cfg = {
            "train": "test/data/toy/train",
            "src": {
                "lang": "de",
                "level": "word",
                "lowercase": False,
                "max_length": 30,
                "voc_limit": self.voc_limit,
            },
            "trg": {
                "lang": "en",
                "level": "word",
                "lowercase": False,
                "max_length": 30,
                "voc_limit": self.voc_limit,
            },
            "dataset_type": "plain",
        }

        self.sents = [
            "Die Wahrheit ist, dass die Titanic – obwohl sie alle Kinokassenrekorde "
            "bricht – nicht gerade die aufregendste Geschichte vom Meer ist.",
            "GROẞ",  # ẞ (uppercase) requires Unicode
        ]
        self.word_list = set((" ".join(self.sents)).split())  # only unique tokens
        self.char_list = set(list(" ".join(self.sents)))  # only unique tokens
        self.vocab_file_bpe = Path("test/data/toy/bpe200.txt")
        self.vocab_file_sp = Path("test/data/toy/sp200.txt")
        self.word_vocab = Vocabulary(tokens=sorted(list(self.word_list)))
        self.char_vocab = Vocabulary(tokens=sorted(list(self.char_list)))
        self.specials = ["<unk>", "<pad>", "<s>", "</s>"]

    def testVocabularyFromList(self):
        self.assertEqual(
            len(self.word_vocab) - len(self.word_vocab.specials),
            len(self.word_list),
        )
        self.assertEqual(
            len(self.char_vocab) - len(self.char_vocab.specials),
            len(self.char_list),
        )
        expected_char_itos = [
            " ",
            ",",
            ".",
            "D",
            "G",
            "K",
            "M",
            "O",
            "R",
            "T",
            "W",
            "a",
            "b",
            "c",
            "d",
            "e",
            "f",
            "g",
            "h",
            "i",
            "k",
            "l",
            "m",
            "n",
            "o",
            "r",
            "s",
            "t",
            "u",
            "v",
            "w",
            "ẞ",
            "–",
        ]

        # pylint: disable=protected-access
        self.assertEqual(self.char_vocab._itos, self.specials + expected_char_itos)
        expected_word_itos = [
            "Die",
            "GROẞ",
            "Geschichte",
            "Kinokassenrekorde",
            "Meer",
            "Titanic",
            "Wahrheit",
            "alle",
            "aufregendste",
            "bricht",
            "dass",
            "die",
            "gerade",
            "ist,",
            "ist.",
            "nicht",
            "obwohl",
            "sie",
            "vom",
            "–",
        ]
        self.assertEqual(self.word_vocab._itos, self.specials + expected_word_itos)
        # pylint: enable=protected-access

    def testVocabularyFromFile(self):
        # write vocabs to file and create new ones from those files
        tmp_file_word = Path("tmp.src.word")
        tmp_file_char = Path("tmp.src.char")
        self.word_vocab.to_file(tmp_file_word)
        self.char_vocab.to_file(tmp_file_char)

        word_vocab = Vocabulary(tokens=read_list_from_file(tmp_file_word))
        char_vocab = Vocabulary(tokens=read_list_from_file(tmp_file_char))
        self.assertEqual(self.word_vocab, word_vocab)
        self.assertEqual(self.char_vocab, char_vocab)

        # delete tmp files
        tmp_file_char.unlink()
        tmp_file_word.unlink()

        # pylint: disable=protected-access
        bpe_vocab = Vocabulary(tokens=read_list_from_file(self.vocab_file_bpe))
        expected_bpe_itos = [
            "t@@", "s@@", "e", "e@@", "d@@", "o@@", "b@@", "g@@", "en", "m@@", "u@@"
        ]
        self.assertEqual(bpe_vocab._itos[:15], self.specials + expected_bpe_itos)

        sp_vocab = Vocabulary(tokens=read_list_from_file(self.vocab_file_sp))
        expected_sp_itos = ["▁", "e", "s", "t", "o", "i", "n", "en", "m", "r", "er"]
        self.assertEqual(sp_vocab._itos[:15], self.specials + expected_sp_itos)
        # pylint: enable=protected-access

    def testVocabularyFromDataset(self):
        src_vocab, trg_vocab, _, _, _ = load_data(self.cfg, datasets=["train"])
        self.assertEqual(len(src_vocab), self.voc_limit + len(self.specials))
        self.assertEqual(len(trg_vocab), self.voc_limit + len(self.specials))

        expected_src_itos = [
            "die",
            "und",
            "der",
            "ist",
            "in",
            "das",
            "wir",
            "zu",
            "Sie",
            "es",
            "von",
        ]
        expected_trg_itos = [
            "the",
            "of",
            "to",
            "and",
            "a",
            "that",
            "in",
            "is",
            "you",
            "we",
            "And",
        ]
        # pylint: disable=protected-access
        self.assertEqual(src_vocab._itos[:15], self.specials + expected_src_itos)
        self.assertEqual(trg_vocab._itos[:15], self.specials + expected_trg_itos)

    def testIsUnk(self):
        self.assertTrue(self.word_vocab.is_unk("BLA"))
        self.assertFalse(self.word_vocab.is_unk("Die"))
        self.assertFalse(self.word_vocab.is_unk("GROẞ"))
        self.assertTrue(self.char_vocab.is_unk("x"))
        self.assertFalse(self.char_vocab.is_unk("d"))
        self.assertFalse(self.char_vocab.is_unk("ẞ"))

    def testEncodingDecoding(self):
        tokenized = [s.split() for s in self.sents]
        ids, length = self.word_vocab.sentences_to_ids(tokenized, bos=True, eos=True)
        expected_ids = [
            [
                2, 4, 10, 17, 14, 15, 9, 23, 20, 21, 11, 7, 13, 23, 19, 16, 15, 12, 6,
                22, 8, 18, 3
            ],
            [2, 5, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        ]
        expected_length = [23, 3]
        self.assertEqual(ids, expected_ids)
        self.assertEqual(length, expected_length)

        decoded = self.word_vocab.arrays_to_sentences(np.array(ids),
                                                      cut_at_eos=True,
                                                      skip_pad=True)
        self.assertEqual(decoded[0], ["<s>"] + tokenized[0] + ["</s>"])
        self.assertEqual(decoded[1], ["<s>"] + tokenized[1] + ["</s>"])



# coding: utf-8
"""
Various encoders
"""
from typing import Tuple

import torch
from torch import Tensor, nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from joeynmt.helpers import freeze_params
from joeynmt.transformer_layers import PositionalEncoding, TransformerEncoderLayer


class Encoder(nn.Module):
    """
    Base encoder class
    """

    # pylint: disable=abstract-method
    @property
    def output_size(self):
        """
        Return the output size

        :return:
        """
        return self._output_size


class RecurrentEncoder(Encoder):
    """Encodes a sequence of word embeddings"""

    # pylint: disable=unused-argument
    def __init__(
        self,
        rnn_type: str = "gru",
        hidden_size: int = 1,
        emb_size: int = 1,
        num_layers: int = 1,
        dropout: float = 0.0,
        emb_dropout: float = 0.0,
        bidirectional: bool = True,
        freeze: bool = False,
        **kwargs,
    ) -> None:
        """
        Create a new recurrent encoder.

        :param rnn_type: RNN type: `gru` or `lstm`.
        :param hidden_size: Size of each RNN.
        :param emb_size: Size of the word embeddings.
        :param num_layers: Number of encoder RNN layers.
        :param dropout:  Is applied between RNN layers.
        :param emb_dropout: Is applied to the RNN input (word embeddings).
        :param bidirectional: Use a bi-directional RNN.
        :param freeze: freeze the parameters of the encoder during training
        :param kwargs:
        """
        super().__init__()

        self.emb_dropout = torch.nn.Dropout(p=emb_dropout, inplace=False)
        self.type = rnn_type
        self.emb_size = emb_size

        rnn = nn.GRU if rnn_type == "gru" else nn.LSTM

        self.rnn = rnn(
            emb_size,
            hidden_size,
            num_layers,
            batch_first=True,
            bidirectional=bidirectional,
            dropout=dropout if num_layers > 1 else 0.0,
        )

        self._output_size = 2 * hidden_size if bidirectional else hidden_size

        if freeze:
            freeze_params(self)

    def _check_shapes_input_forward(self, embed_src: Tensor, src_length: Tensor,
                                    mask: Tensor) -> None:
        """
        Make sure the shape of the inputs to `self.forward` are correct.
        Same input semantics as `self.forward`.

        :param embed_src: embedded source tokens
        :param src_length: source length
        :param mask: source mask
        """
        # pylint: disable=unused-argument
        assert embed_src.shape[0] == src_length.shape[0]
        assert embed_src.shape[2] == self.emb_size
        # assert mask.shape == embed_src.shape
        assert len(src_length.shape) == 1

    def forward(self, embed_src: Tensor, src_length: Tensor, mask: Tensor,
                **kwargs) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Applies a bidirectional RNN to sequence of embeddings x.
        The input mini-batch x needs to be sorted by src length.
        x and mask should have the same dimensions [batch, time, dim].

        :param embed_src: embedded src inputs,
            shape (batch_size, src_len, embed_size)
        :param src_length: length of src inputs
            (counting tokens before padding), shape (batch_size)
        :param mask: indicates padding areas (zeros where padding), shape
            (batch_size, src_len, embed_size)
        :param kwargs:
        :return:
            - output: hidden states with
                shape (batch_size, max_length, directions*hidden),
            - hidden_concat: last hidden state with
                shape (batch_size, directions*hidden)
        """
        self._check_shapes_input_forward(embed_src=embed_src,
                                         src_length=src_length,
                                         mask=mask)
        total_length = embed_src.size(1)

        # apply dropout to the rnn input
        embed_src = self.emb_dropout(embed_src)

        packed = pack_padded_sequence(embed_src, src_length.cpu(), batch_first=True)
        output, hidden = self.rnn(packed)

        if isinstance(hidden, tuple):
            hidden, memory_cell = hidden  # pylint: disable=unused-variable

        output, _ = pad_packed_sequence(output,
                                        batch_first=True,
                                        total_length=total_length)
        # hidden: dir*layers x batch x hidden
        # output: batch x max_length x directions*hidden
        batch_size = hidden.size()[1]
        # separate final hidden states by layer and direction
        hidden_layerwise = hidden.view(
            self.rnn.num_layers,
            2 if self.rnn.bidirectional else 1,
            batch_size,
            self.rnn.hidden_size,
        )
        # final_layers: layers x directions x batch x hidden

        # concatenate the final states of the last layer for each directions
        # thanks to pack_padded_sequence final states don't include padding
        fwd_hidden_last = hidden_layerwise[-1:, 0]
        bwd_hidden_last = hidden_layerwise[-1:, 1]

        # only feed the final state of the top-most layer to the decoder
        # pylint: disable=no-member
        hidden_concat = torch.cat([fwd_hidden_last, bwd_hidden_last], dim=2).squeeze(0)
        # final: batch x directions*hidden

        assert hidden_concat.size(0) == output.size(0), (
            hidden_concat.size(),
            output.size(),
        )
        return output, hidden_concat

    def __repr__(self):
        return f"{self.__class__.__name__}(rnn={self.rnn})"


class TransformerEncoder(Encoder):
    """
    Transformer Encoder
    """

    def __init__(
        self,
        hidden_size: int = 512,
        ff_size: int = 2048,
        num_layers: int = 8,
        num_heads: int = 4,
        dropout: float = 0.1,
        emb_dropout: float = 0.1,
        freeze: bool = False,
        **kwargs,
    ):
        """
        Initializes the Transformer.
        :param hidden_size: hidden size and size of embeddings
        :param ff_size: position-wise feed-forward layer size.
          (Typically this is 2*hidden_size.)
        :param num_layers: number of layers
        :param num_heads: number of heads for multi-headed attention
        :param dropout: dropout probability for Transformer layers
        :param emb_dropout: Is applied to the input (word embeddings).
        :param freeze: freeze the parameters of the encoder during training
        :param kwargs:
        """
        super().__init__()

        self._output_size = hidden_size

        # build all (num_layers) layers
        self.layers = nn.ModuleList([
            TransformerEncoderLayer(
                size=hidden_size,
                ff_size=ff_size,
                num_heads=num_heads,
                dropout=dropout,
                alpha=kwargs.get("alpha", 1.0),
                layer_norm=kwargs.get("layer_norm", "post"),
            ) for _ in range(num_layers)
        ])

        self.pe = PositionalEncoding(hidden_size)
        self.emb_dropout = nn.Dropout(p=emb_dropout)

        self.layer_norm = (nn.LayerNorm(hidden_size, eps=1e-6) if kwargs.get(
            "layer_norm", "post") == "pre" else None)

        if freeze:
            freeze_params(self)

    def forward(
        self,
        embed_src: Tensor,
        src_length: Tensor,  # unused
        mask: Tensor = None,
        **kwargs,
    ) -> Tuple[Tensor, Tensor]:
        """
        Pass the input (and mask) through each layer in turn.
        Applies a Transformer encoder to sequence of embeddings x.
        The input mini-batch x needs to be sorted by src length.
        x and mask should have the same dimensions [batch, time, dim].

        :param embed_src: embedded src inputs,
            shape (batch_size, src_len, embed_size)
        :param src_length: length of src inputs
            (counting tokens before padding), shape (batch_size)
        :param mask: indicates padding areas (zeros where padding), shape
            (batch_size, 1, src_len)
        :return:
            - output: hidden states with shape (batch_size, max_length, hidden)
            - None
        """
        # pylint: disable=unused-argument
        x = self.pe(embed_src)  # add position encoding to word embeddings
        x = self.emb_dropout(x)

        for layer in self.layers:
            x = layer(x, mask)

        if self.layer_norm is not None:
            x = self.layer_norm(x)
        return x, None

    def __repr__(self):
        return (f"{self.__class__.__name__}(num_layers={len(self.layers)}, "
                f"num_heads={self.layers[0].src_src_att.num_heads}, "
                f"alpha={self.layers[0].alpha}, "
                f'layer_norm="{self.layers[0]._layer_norm_position}")')



from test.unit.test_helpers import TensorTestCase

import torch

from joeynmt.transformer_layers import MultiHeadedAttention, PositionalEncoding


class TestTransformerUtils(TensorTestCase):

    def setUp(self):
        seed = 42
        torch.manual_seed(seed)

    def test_position_encoding(self):
        batch_size = 2
        max_time = 3
        emb_size = hidden_size = 12

        x = torch.zeros([batch_size, max_time, emb_size])
        pe = PositionalEncoding(emb_size)
        output = pe(x)
        self.assertEqual(pe.pe.size(2), hidden_size)
        self.assertTensorAlmostEqual(output, pe.pe[:, :x.size(1)])

    def test_src_src_attention(self):
        num_heads = 2
        hidden_size = 8
        att = MultiHeadedAttention(num_heads=num_heads, size=hidden_size)

        batch_size = 1
        src_len = 4
        x = torch.randn(batch_size, src_len, hidden_size)

        output, attention_weights = att(x, x, x)
        self.assertEqual(output.size(), (batch_size, src_len, hidden_size))
        self.assertIsNone(attention_weights)  # return_weights = False by default

    def test_src_trg_attention(self):
        num_heads = 2
        hidden_size = 8
        att = MultiHeadedAttention(num_heads=num_heads, size=hidden_size)

        batch_size = 1
        src_len = 4
        trg_len = 5
        x = torch.randn(batch_size, src_len, hidden_size)
        m = torch.randn(batch_size, trg_len, hidden_size)

        output, attention_weights = att(x, x, m, return_weights=True)
        self.assertEqual(output.size(), (batch_size, trg_len, hidden_size))
        self.assertEqual(attention_weights.size(), (batch_size, trg_len, src_len))
